#include "Command.h"
#include "DataStructures.h"

/////////////////////////////////////////////////////////////////////
// Each of the following sets of Execute/Unexecute commands
// correspond to an "undoable" action.


// !!! Some undo operations are the equivalent of do operations
//        ie do AddPoint and undo RemovePoint


/////////////////////////////////////////////////////////////////////
// Resize, Un-resize the diameter of a root
bool UC_ResizeDiameter::Execute()
{
  ApplicationData& app = m_data->GetApplicationData();
  Experiment& exp = m_data->GetExperiment();
  Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
  if(tube == NULL) return false;
  Root *root = tube->GetRoot( m_root_id );
  if(root == NULL) return false;
  RootInstance *root_inst = root->GetRootInstance(m_session);
  if(root_inst == NULL) return false;
  
  return root_inst->ResizeDiameter(m_index, m_new_radius);
}

bool UC_ResizeDiameter::Unexecute()
{
  ApplicationData& app = m_data->GetApplicationData();
  Experiment& exp = m_data->GetExperiment();
  Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
  if(tube == NULL) return false;
  Root *root = tube->GetRoot( m_root_id );
  if(root == NULL) return false;
  RootInstance *root_inst = root->GetRootInstance(m_session);
  if(root_inst == NULL) return false;
  
  return root_inst->ResizeDiameter(m_index, m_old_radius);
}

/////////////////////////////////////////////////////////////////////
// Set, Unset image annotated flag
bool UC_SetImageAnnotated::Execute()
{
  ApplicationData& app = m_data->GetApplicationData();
  Experiment& exp = m_data->GetExperiment();
  Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
  if (tube == NULL) return false;
  
  if (tube->GetImageAnnotated(m_window, m_session) != m_annotated)
  {
    tube->SetImageAnnotated(true, m_window, m_session);
    return true;
  }
  return false;
}

bool UC_SetImageAnnotated::Unexecute()
{
  ApplicationData& app = m_data->GetApplicationData();
  Experiment& exp = m_data->GetExperiment();
  Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
  if (tube == NULL) return false;
  
  if (tube->GetImageAnnotated(m_window, m_session) == m_annotated)
  {
    tube->SetImageAnnotated(false, m_window, m_session);
    return true;
  }
  return false;
}

/////////////////////////////////////////////////////////////////////
// Set, Unset root alive flag
//
// When setting root instances to alive, all current and past sessions
// of the root instance are set to alive.
// When setting root instances to dead, all current and future sessions
// of the root instance are set to dead
bool UC_SetRootAlive::Execute()
{
  ApplicationData& app = m_data->GetApplicationData();
  Experiment& exp = m_data->GetExperiment();
  int session = m_session;
  
  Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
  if(tube == NULL) return false;
  Root *root = tube->GetRoot( m_rootid );
  if(root == NULL) return false;
  RootInstance *inst = root->GetRootInstance(m_session);
  if(inst == NULL) return false;
  
  // m_alive is what we are setting the root instance to
  if(m_alive)
  {       
    // set root instance to alive and in the current and past sessions
    while(session > 0)
    {
      inst = root->GetRootInstance(session);
      if (inst != NULL && !(inst->IsAlive()))  inst->SetAlive();
      session--;
    }
  } 
  else 
  {
    // set root instance to dead in the current and future sessions
    while(session <= exp.GetMaxSessions())
    {
      inst = root->GetRootInstance(session);
      if (inst != NULL && inst->IsAlive())     inst->SetDead();
      session++;
    }
  }  
  
  return true;
}

// When undoing a set root instances to alive, all current and past 
// sessions of the root instance are set to dead
// When undoing a set root instances to dead, all current and future 
// sessions of the root instance are set to alive
bool UC_SetRootAlive::Unexecute()
{
  ApplicationData& app = m_data->GetApplicationData();
  Experiment& exp = m_data->GetExperiment();
  int session = m_session;
  
  Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
  if(tube == NULL) return false;
  Root *root = tube->GetRoot( m_rootid );
  if(root == NULL) return false;
  RootInstance *inst = root->GetRootInstance(m_session);
  if(inst == NULL) return false;
  
  // m_alive is what we set the root instance to, undo by setting to !m_alive
  if(m_alive)
  {       
    // set root instances to dead in current and future sessions
    while(session <= exp.GetMaxSessions())
    {
      inst = root->GetRootInstance(session);
      if (inst != NULL && inst->IsAlive())   inst->SetDead();
      session++;
    }
  } 
  else 
  {
    // set root instances to alive in current and past sessions
    while(session > 0)
    {
      inst = root->GetRootInstance(session);
      if (inst != NULL && !inst->IsAlive())  inst->SetAlive();
      session--;    
    }
  }  
  
  return true;
}

/////////////////////////////////////////////////////////////////////
// Move, Un-move a single root in a session in two dimensional 
// space by a pair of offsets
bool UC_MoveRootSess::Execute()
{
  // if this is a Redo action, execute the following move actions
  if(m_executed_once)
  {
    m_root_inst = RootSessGetRootInstance( m_root_id );
    if( m_root_inst == NULL ) return false;
    
    // Move each point on the root to its original location using x and y offsets 
    for(int i = 0;i<m_root_inst->GetAllPoints()->size();i++)
    {
      //may need to update this to be a bit different with x and y
      RootSessMovePoint(m_root_inst, i, m_root_id, m_root_inst->GetPoint(i)->x+m_offset_x,m_root_inst->GetPoint(i)->y+m_offset_y);
    }
    
    // Move Diameter (assumes only one diameter)
    RootSessMoveDiameter(m_root_inst, 0,m_root_id, m_root_inst->GetDiameter(0)->r_point.x + m_offset_x, m_root_inst->GetDiameter(0)->r_point.y + m_offset_y);
  }
  else m_executed_once = true; // Is this needed? Does this need to be undone below?
  
  return true;
}

bool UC_MoveRootSess::Unexecute()
{
  m_root_inst = RootSessGetRootInstance( m_root_id );
  if( m_root_inst == NULL ) return false;
  
  // Move each point on the root to its original location using x and y offsets 
  for(int i = 0;i<m_root_inst->GetAllPoints()->size();i++)
  {
    //may need to update this to be a bit different with x and y
    RootSessMovePoint(m_root_inst, i, m_root_id, m_root_inst->GetPoint(i)->x-m_offset_x,m_root_inst->GetPoint(i)->y-m_offset_y);
  }
  
  /// Move Diameter (assumes only one diameter)
  RootSessMoveDiameter(m_root_inst, 0,m_root_id, m_root_inst->GetDiameter(0)->r_point.x - m_offset_x, m_root_inst->GetDiameter(0)->r_point.y - m_offset_y);
  
  return true;
}


/////////////////////////////////////////////////////////////////////
// Move, Move-back the diameter of a root in two dimensional 
// space by a pair of offsets
bool UC_MoveDiameter::Execute()
{
  m_root_inst = MoveDiameterGetRootInstance(m_root_id);
  if(m_root_inst == NULL) return false;
  
  return ((m_root_inst->MoveDiameter(m_index, m_new_pt.x, m_new_pt.y)) 
    && (m_root_inst->ChangeWindowOfDiameter(m_index, m_new_pt.window)));
}

bool UC_MoveDiameter::Unexecute()
{
  m_root_inst = MoveDiameterGetRootInstance(m_root_id);
  if(m_root_inst == NULL) return false;
  
  return ((m_root_inst->MoveDiameter(m_index, m_old_pt.x, m_old_pt.y))
    && (m_root_inst->ChangeWindowOfDiameter(m_index, m_old_pt.window)));
}


/////////////////////////////////////////////////////////////////////
// Move, Move-back all of the roots in a window in two dimensional 
// space by a pair of offsets
bool UC_MoveRootsInWindow::Execute()
{
  if(m_executed_once)
  {
    Experiment& exp = m_data->GetExperiment();  
    Tube *tube = exp.GetTubeByName(m_data->GetApplicationData().GetCurrentTubeName());
    if(tube == NULL) return false;
    vector<Root> *roots = tube->GetRootsInWindow(m_window);
    if (roots == NULL) return false;
    
    vector<Root>::iterator iter;
    for(iter = roots->begin();iter != roots->end();iter++)
    {
      RootWindowMoveRootSess(iter->GetID(), 
        m_session, m_offset_x, m_offset_y);
    }
  }
  
  else m_executed_once = true; // Is this needed? Does this need to be undone below?
  
  return true;
}

bool UC_MoveRootsInWindow::Unexecute()
{
  Experiment& exp = m_data->GetExperiment();  
  Tube *tube = exp.GetTubeByName(m_data->GetApplicationData().GetCurrentTubeName());
  if(tube == NULL) return false;
  vector<Root> *roots = tube->GetRootsInWindow(m_window);
  if (roots == NULL) return false;
  
  vector<Root>::iterator iter;
  for(iter = roots->begin();iter != roots->end();iter++)
  {
    RootWindowMoveRootSess(iter->GetID(), 
      m_session, -m_offset_x, -m_offset_y);
  }
  
  return true;
}


/////////////////////////////////////////////////////////////////////
// Set, Unset the color of a root
bool UC_SetRootColor::Execute() 
{
  ApplicationData& app = m_data->GetApplicationData();
  Experiment& exp = m_data->GetExperiment();
  
  Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
  if(tube == NULL) return false;
  Root *root = tube->GetRoot( m_root_id );
  if(root == NULL) return false;
  RootInstance *inst = root->GetRootInstance(m_session);
  if(inst == NULL) return false;
  
  m_old_color = inst->GetColor();
  if(m_old_color != m_color_choice)
  {
    /// Set status unsaved
    inst->SetColor(m_color_choice);
    return true;
  }
  return false;
}

bool UC_SetRootColor::Unexecute() 
{
  ApplicationData& app = m_data->GetApplicationData();
  Experiment& exp = m_data->GetExperiment();
  
  Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
  if(tube == NULL) return false;
  Root *root = tube->GetRoot( m_root_id );
  if(root == NULL) return false;
  RootInstance *inst = root->GetRootInstance(m_session);
  if(inst == NULL) return false;
  
  inst->SetColor(m_old_color);
  return true;
}


/////////////////////////////////////////////////////////////////////
// Move, Move-back a point in a root 
bool UC_MovePoint::Execute()
{
  m_root_inst = MovePointGetRootInstance(m_root_id);
  if(m_root_inst == NULL) return false;
  
  m_root_inst->MoveDiameter(0, m_new_diam.r_point.x, m_new_diam.r_point.y);
  m_root_inst->ChangeWindowOfPoint(m_index, m_new_pt.window);
  
  if(m_root_inst->MovePoint(m_index, m_new_pt.x, m_new_pt.y)) return true;
  return false;
}

bool UC_MovePoint::Unexecute()
{
  m_root_inst = MovePointGetRootInstance(m_root_id);
  if(m_root_inst == NULL) return false;
  
  m_root_inst->MoveDiameter(0, m_old_diam.r_point.x, m_old_diam.r_point.y); 
  m_root_inst->ChangeWindowOfPoint(m_index, m_old_pt.window);
  
  if(m_root_inst->MovePoint(m_index, m_old_pt.x, m_old_pt.y)) return true;
  return false;
}

/////////////////////////////////////////////////////////////////////
// Add, Un-add a point to a root instance
bool UC_AddPoint::Execute()
{
  m_root_inst = AddPointGetRootInstance();
  if(m_root_inst == NULL) return false;
  
  m_point = m_root_inst->InsertPoint(m_index);
  m_diameter = m_root_inst->GetDiameter(0);
  
  // save the old locations
  int temp_diam_x = m_diameter->r_point.x;
  int temp_diam_y = m_diameter->r_point.y;
  
  if(m_temp_point.x > 0 && m_temp_point.y > 0 && m_temp_point.window > 0)
  {
    // assign coordinates to the new point using the previous location
    m_point->x = m_temp_point.x;
    m_point->y = m_temp_point.y;
    m_point->window = m_temp_point.window;
    
    // move the diameter to the previous location, stored in the m_diam variables
    m_root_inst->MoveDiameter(0, m_diam_x, m_diam_y);
  }
  
  // to Undo this, we will need the original location in Unexecute
  // store this location in the m_diam variables
  m_diam_x = temp_diam_x;
  m_diam_y = temp_diam_y;
  return true;
}

bool UC_AddPoint::Unexecute()
{
  m_root_inst = AddPointGetRootInstance();
  if(m_root_inst == NULL) return false;
  
  m_point = m_root_inst->GetPoint(m_index+1);
  m_temp_point.x = m_point->x;
  m_temp_point.y = m_point->y;
  m_temp_point.window = m_point->window;
  
  m_root_inst->DeletePoint(m_index+1);
  
  // save the old location of the diameter, then move it back
  m_diameter = m_root_inst->GetDiameter(0);
  int temp_diam_x = m_diameter->r_point.x;
  int temp_diam_y = m_diameter->r_point.y;
  m_root_inst->MoveDiameter(0, m_diam_x, m_diam_y); 
  
  // remember the location of the diameter for the
  // Redo operation. store in the m_diam variables.
  m_diam_x = temp_diam_x;
  m_diam_y = temp_diam_y;
  return true;
}


/////////////////////////////////////////////////////////////////////
// Delete, undelete a point
bool UC_DeletePoint::Execute()
{
  m_root_inst = DeletePointGetRootInstance(m_root_id);
  if(m_root_inst == NULL) return false;
  
  // save the location of the diameter so delete point can be undone
  m_diameter = m_root_inst->GetDiameter(0);
  int temp_diam_x = m_diameter->r_point.x;
  int temp_diam_y = m_diameter->r_point.y;
  
  // move diameter to original position. applicable only for a Redo
  // operation. the if statement checks if this is a Redo
  if(m_temp_point.x > 0 && m_temp_point.y > 0 && m_temp_point.window > 0)
  {
    m_root_inst->MoveDiameter(0, m_diam_x, m_diam_y);
  }
  m_diam_x = temp_diam_x;
  m_diam_y = temp_diam_y;
  
  // store point coordinates for the Undo operation
  m_point = m_root_inst->GetPoint(m_index);
  m_temp_point.x = m_point->x;
  m_temp_point.y = m_point->y;
  m_temp_point.window = m_point->window;
  
  m_root_inst->DeletePoint(m_index);
  
  // Note: DeletePoint() from DataStructures.cpp returns TRUE
  // if the erase operator from Vector returns NULL. But erase
  // returns the new location of the next element, which is not
  // supposed to be NULL. - Ninad Pradhan (04/24/08)
  
  return true;
}

bool UC_DeletePoint::Unexecute()
{
  m_root_inst = DeletePointGetRootInstance(m_root_id);
  if(m_root_inst == NULL) return false;
  
  // create a new point with the deleted point's location
  m_point = m_root_inst->InsertPoint(m_index-1);
  m_point->x = m_temp_point.x;
  m_point->y = m_temp_point.y;
  m_point->window = m_temp_point.window;
  
  // save the location of the diameter so we can undo the move
  m_diameter = m_root_inst->GetDiameter(0);
  int temp_diam_x = m_diameter->r_point.x;
  int temp_diam_y = m_diameter->r_point.y;
  m_root_inst->MoveDiameter(0, m_diam_x, m_diam_y);
  m_diam_x = temp_diam_x;
  m_diam_y = temp_diam_y;
  
  return true;
}


/////////////////////////////////////////////////////////////////////
// Create, uncreate a root
bool UC_CreateRoot::Execute() 
{
  Experiment& exp = m_data->GetExperiment();
  Tube* tube = exp.GetTubeByName(m_tube_tag);
  Root* temp_root = NULL;
  
  if (tube == NULL)  tube = exp.CreateTube(m_tube_tag);
  assert(tube);
  
  if(!m_executed_once)
  {
    temp_root = tube->CreateRoot(m_primary_window);
    m_root_id = temp_root->GetID();
    m_executed_once = true;
  }
  else
  {
    temp_root = tube->CreateRoot(m_primary_window, m_root_id);
  }
  //gets session number from root instance because may not be drawing in current session. (3x3 grid) -MTB
  int session = m_ri_orig.GetSessionNumber();
  
  assert(temp_root);
  RootInstance* ri = temp_root->CreateRootInstance(session);
  
  *ri = m_ri_orig;
  assert(ri);
  ri->SetConditionCode(m_default_cond);
  //Automatically set to annoated when created outside middle window for 3x3 grid -MTB
  tube->SetImageAnnotated(true,m_primary_window,session);
  
  return true;
}

bool UC_CreateRoot::Unexecute() 
{
  
  Experiment& exp = m_data->GetExperiment();
  Tube* tube = exp.GetTubeByName(m_tube_tag);
  
  /// If tube does not exist, create a new tube
  if (tube == NULL)  tube = exp.CreateTube(m_tube_tag);
  assert(tube);
  
  
  return tube->DeleteRoot(m_root_id);
  
}

/////////////////////////////////////////////////////////////////////
// Delete, undelete a root
bool UC_DeleteRoot::Execute() 
{
  Experiment& exp = m_data->GetExperiment();
  Tube* tube = exp.GetTubeByName(m_tube_tag);
  if(tube == NULL) return false;
  
  return tube->DeleteRoot(m_root_id);
}

bool UC_DeleteRoot::Unexecute() 
{
  Experiment& exp = m_data->GetExperiment();
  Tube* tube = exp.GetTubeByName(m_tube_tag);
  Root* new_root = NULL;
  
  // If tube does not exist, create a new tube
  if (tube == NULL)  tube = exp.CreateTube(m_tube_tag);
  assert(tube);
  
  new_root = tube->CreateRoot(m_primary_window, m_root_id);
  
  *new_root = m_root_orig;
  
  return true;
}

/////////////////////////////////////////////////////////////////////
// Create, uncreate a point in a root instance
bool UC_CreatePoint::Execute()
{
  if(m_root_inst == NULL) return false;
  
  m_temp_point = m_root_inst->CreatePoint();
  return true;
}

bool UC_CreatePoint::Unexecute()
{
  if(m_root_inst == NULL) return false;
  
  vector<RPoint>* points = m_root_inst->GetAllPoints();
  if (points == NULL) return false;
  
  vector<RPoint>::iterator iter;
  for(iter = points->begin(); iter != points->end(); iter++)
  {
    if ((*iter) == (*m_temp_point)) return m_root_inst->DeletePoint(iter - points->begin());
  }
  return false;
}

/////////////////////////////////////////////////////////////////////
// Copy, uncopy a session (for C_CopySession)
bool UC_CopySession::Execute()
{
  Experiment& exp = m_data->GetExperiment();
  Tube *tube = exp.GetTubeByName(m_tubename);
  if(tube == NULL)  return false;
  
  vector<Tube> *alltubes;
  vector<Root> *roots;
  const vector<RootInstance> *instances;
  bool copy_all = (m_curr_window == -1);
  CString str;
  DWORD choice;
  
  bool notified=false;
  int ok_to_copy = true;    // This is set to false later only if user cancels operation based on message box
  int i,j,k,m;
  
  //Check the bounds of the source and destination sessions
  assert(m_from >= 1 && m_to >= 1);
  if (m_from < 1 || m_to < 1) 
  {
    AfxMessageBox("Source and destination sessions must be positive.  Unable to copy.", MB_ICONINFORMATION);
    return false;
  }
  assert(m_from + 1 <= m_to);
  if (m_from + 1 > m_to) 
  {
    AfxMessageBox("Source session must be less than destination session.  Unable to copy.", MB_ICONINFORMATION);
    return false;
  }
  if (m_to > exp.GetMaxSessions())
  {
    str.Format("Destination session is too large.  Unable to copy.\r\n\r\nDestination session must be less than or equal to max session with images.\r\nSpecified Destination Session = %d\r\nMax Session with Images = %d\r\n",m_to,exp.GetMaxSessions());
    AfxMessageBox(str, MB_ICONINFORMATION);
    return false;
  }
  
  alltubes = exp.GetAllTubes();
  
  //Check to see if roots exist in destination sessions.  If they are, prompt the user to see
  //if they want them overwritten
  for (i=0; i<alltubes->size(); i++) 
  {
    roots = alltubes->at(i).GetAllRoots();
    m_roots = alltubes->at(i).GetAllRoots();
    
    //Loop through ALL roots in the current tube.
    for (j=0; j<roots->size(); j++)
    {
      if(copy_all || (roots->at(j).GetPrimaryWindow() == m_curr_window && tube->GetTag() == alltubes->at(i).GetTag()))
      {
        instances = roots->at(j).GetAllRootInstances();
        //Loop through ALL sessions in the current root to determine if there already exists a session
        //in the selected session interval
        for( k=0; k<instances->size(); k++)
        {
          for( m= m_from+1; m <= m_to; m++)
          {
            // If the session number of this root instance is equal to l, which is in the interval from "from" to "to"
            //  ask the user if they want to overwrite the destination sessions.
            if( instances->at(k).GetSessionNumber() == m && !notified)
            {
              if (copy_all)
              {
                if (m_from+1 == m_to)
                {
                  str.Format("Session %d already contains roots.\r\nAre you sure you want to overwrite this data?", m_to);
                }
                else
                {
                  str.Format("Sessions %d through %d already contain roots.\r\nAre you sure you want to overwrite this data?", m_from+1, m_to);
                }
              }
              else
              {
                if (m_from+1 == m_to)
                {
                  str.Format("Window %d of Session %d already contains roots.\r\nAre you sure you want to overwrite this data?", m_curr_window, m_to);
                }
                else
                {
                  str.Format("Window %d of Sessions %d through %d already contain roots.\r\nAre you sure you want to overwrite this data?", m_curr_window, m_from+1, m_to);
                }
              }
              choice = AfxMessageBox(str, MB_YESNO | MB_DEFBUTTON2);
              if (choice != IDYES)  ok_to_copy = false;
              notified = true;
            }
          }
        }
      }
    }
  }
  
  
  //If the user chose to overwrite (or wasn't prompted at all) continue
  if (ok_to_copy) 
  {
    //Copy over all tubes and windows
    if (copy_all)
    {
      for (i=0; i<alltubes->size();i++)
      {
        roots = alltubes->at(i).GetAllRoots();
        for (j=0; j < roots->size(); j++) 
        {
          for (k=m_from+1; k <=m_to; k++) 
          {
            roots->at(j).DeleteRootInstance(k);
            roots->at(j).CopyRootInstance(m_from, k);
            //Delete entire root if last instance is deleted
            if (roots->at(j).IsEmpty()) {
              alltubes->at(i).DeleteRoot(roots->at(j).GetID());
            }
          }
        }
      }      
    }
    //else only copy within the current window
    else 
    {
      roots = tube->GetAllRoots();
      for(int i = 0; i < roots->size(); i++)
      {
        if ( m_which_root == -1 || roots->at(i).GetDisplayableID() == m_which_root)
        {
          if(roots->at(i).GetPrimaryWindow() == m_curr_window)
          {   
            for (k=m_from+1; k<=m_to;k++) 
            {
              roots->at(i).DeleteRootInstance(k);
              roots->at(i).CopyRootInstance(m_from, k);
              //Delete entire root if last instance is deleted
              if (roots->at(i).IsEmpty()) {
                tube->DeleteRoot(roots->at(i).GetID());
              }
            }
          }
        }
      }
    }
  }
  return true;
}


bool UC_CopySession::Unexecute()
{
  Experiment& exp = m_data->GetExperiment();
  Tube *tube = exp.GetTubeByName(m_tubename);
  vector<Tube> *alltubes;
  vector<Root> *roots;
  const vector<RootInstance> *instances;
  bool copy_all = (m_curr_window == -1);
  
  bool notified=false;
  int ok_to_copy = true;    // This is set to false later only if user cancels operation based on message box
  int i,j,k,m;
  
  /// Failure if tube is null
  if(tube == NULL)  return false;
  
  //Check the bounds of the source and destination sessions
  assert(m_from >= 1 && m_to >= 1);
  if (m_from < 1 || m_to < 1) 
  {
    AfxMessageBox("Source and destination sessions must be positive.  Unable to copy.", MB_ICONINFORMATION);
    return false;
  }
  assert(m_from + 1 <= m_to);
  if (m_from + 1 > m_to) 
  {
    AfxMessageBox("Source session must be less than destination session.  Unable to copy.", MB_ICONINFORMATION);
    return false;
  }
  
  alltubes = exp.GetAllTubes();
  
  
  //Check to see if roots exist in destination sessions.  If they are, prompt the user to see
  //if they want them overwritten
  for (i=0; i<alltubes->size(); i++) 
  {
    roots = alltubes->at(i).GetAllRoots();
    m_roots = alltubes->at(i).GetAllRoots();
    
    //Loop through ALL roots in the current tube.
    for (j=0; j<roots->size(); j++)
    {
      if(copy_all || (roots->at(j).GetPrimaryWindow() == m_curr_window && tube->GetTag() == alltubes->at(i).GetTag()))
      {
        instances = roots->at(j).GetAllRootInstances();
        //m_roots = roots->at(j).GetAllRootInstances();
        //Loop through ALL sessions in the current root to determine if there already exists a session
        //in the selected session interval
        for( k=0; k<instances->size(); k++)
        {
          for( m= m_from+1; m <= m_to; m++)
          {
            // If the session number of this root instance is equal to l, which is in the interval from "from" to "to"
            //  ask the user if they want to overwrite the destination sessions.
            if( instances->at(k).GetSessionNumber() == m && !notified)
            {
              CString str;
              if (copy_all)
              {
                if (m_from+1 == m_to)
                {
                  str.Format("Session %d already contains roots.\r\nAre you sure you want to overwrite this data?", m_to);
                }
                else
                {
                  str.Format("Sessions %d through %d already contain roots.\r\nAre you sure you want to overwrite this data?", m_from+1, m_to);
                }
              }
              else
              {
                if (m_from+1 == m_to)
                {
                  str.Format("Window %d of Session %d already contains roots.\r\nAre you sure you want to overwrite this data?", m_curr_window, m_to);
                }
                else
                {
                  str.Format("Window %d of Sessions %d through %d already contain roots.\r\nAre you sure you want to overwrite this data?", m_curr_window, m_from+1, m_to);
                }
              }
              //DWORD choice = AfxMessageBox(str, MB_YESNO | MB_DEFBUTTON2);
              //if (choice != IDYES)  ok_to_copy = false;
              notified = true;
            }
          }
        }
      }
    }
  }
  
  
  //If the user chose to overwrite (or wasn't prompted at all) continue
  if (ok_to_copy) 
  {
    //Copy over all tubes and windows
    if (copy_all)
    {
      for (i=0; i<alltubes->size();i++)
      {
        roots = alltubes->at(i).GetAllRoots();
        for (j=0; j < roots->size(); j++) 
        {
          for (k=m_from+1; k <=m_to; k++) 
          {
            roots->at(j).DeleteRootInstance(k);
            m_roots->at(j).CopyRootInstance(k,m_from);
            //Delete entire root if last instance is deleted
            if (roots->at(j).IsEmpty()) {
              alltubes->at(i).DeleteRoot(roots->at(j).GetID());
            }
          }
        }
      }      
    }
    //else only copy within the current window
    else 
    {
      roots = tube->GetAllRoots();
      for(int i = 0; i < roots->size(); i++)
      {
        if ( m_which_root == -1 || roots->at(i).GetDisplayableID() == m_which_root)
        {
          if(roots->at(i).GetPrimaryWindow() == m_curr_window)
          {   
            for (k=m_from+1; k<=m_to;k++) 
            {
              roots->at(i).DeleteRootInstance(k);
              m_roots->at(i).CopyRootInstance(k,m_from);
              //Delete entire root if last instance is deleted
              if (roots->at(i).IsEmpty()) {
                tube->DeleteRoot(roots->at(i).GetID());
              }
            }

          }
        }
      }
    }
  }
  return true;
}


/////////////////////////////////////////////////////////////////////
// Extend, unextend the endpoint of a root
bool UC_ExtendEndpoint::Execute()
{
  if(m_executed_once)
  {
    ApplicationData& app_data = m_data->GetApplicationData();
    CString tube_tag = app_data.GetCurrentTubeName();
    
    Experiment& exp = m_data->GetExperiment();
    Tube* tube = exp.GetTubeByName(tube_tag);
    if(tube == NULL) return NULL;
    
    Root* temp_root = tube->GetRoot(m_root_id);
    assert(temp_root);
    RootInstance* curr_ri = temp_root->GetRootInstance(m_session);
    *curr_ri = m_new_ri;
  }
  
  else m_executed_once = true; // Is this needed? Does this need to be undone below?
  
  
  return true;
}

bool UC_ExtendEndpoint::Unexecute()
{
  ApplicationData& app_data = m_data->GetApplicationData();
  CString tube_tag = app_data.GetCurrentTubeName();
  
  Experiment& exp = m_data->GetExperiment();
  Tube* tube = exp.GetTubeByName(tube_tag);
  if(tube == NULL)  return NULL;
  
  Root* temp_root = tube->GetRoot(m_root_id);
  assert(temp_root);
  RootInstance* curr_ri = temp_root->GetRootInstance(m_session);
  m_new_ri = *curr_ri;
  *curr_ri = m_old_ri;
  
  return true;
}


/////////////////////////////////////////////////////////////////////
// Set, unset the condition code
bool UC_SetConditionCode::Execute()
{
  ApplicationData& app = m_data->GetApplicationData();
  Experiment& exp = m_data->GetExperiment();
  Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
  if(tube == NULL) return false;
  Root *root = tube->GetRoot( m_root_id );
  if(root == NULL) return false;
  RootInstance *inst = root->GetRootInstance(m_session);
  if(inst == NULL) return false;
  m_prev_condition = inst->GetConditionCode();
  
  // If the condition code will changed
  if(m_prev_condition != m_condition)
  {
    inst->SetConditionCode(m_condition);
    return true;
  }
  return false;
}

bool UC_SetConditionCode::Unexecute()
{
  ApplicationData& app = m_data->GetApplicationData();
  Experiment& exp = m_data->GetExperiment();
  Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
  if(tube == NULL) return false;
  Root *root = tube->GetRoot( m_root_id );
  if(root == NULL) return false;
  RootInstance *inst = root->GetRootInstance(m_session);
  if(inst == NULL) return false;
  
  inst->SetConditionCode(m_prev_condition);
  return true;
}