/* 
* Rootfly:  A program for analyzing and measuring roots. 
* Copyright (c) 2007 Clemson University.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or (at
* your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

// Initial controller implementation

#pragma warning(disable:4786)  //Added to remove compiler warning


#include "Controller.h"

extern DebugLogFile g_debug_logfile;

/**
@brief Attaches an Observer to the Subject
@params o - The observer to be attached
*/
void Subject::Attach (Observer* o) {
  m_observers.push_back(o);
}

/**
@brief Detaches an observer from the subject
@params o - The observer to be detached
*/
void Subject::Detach (Observer* o) {
  int i;
  
  /// Find the observer to detach
  for (i = 0; i < m_observers.size(); i++)
  {
    if(m_observers[i] == o)
      break;
  }
  
  /// Remove the observer from the vector so it will not recieve notifications
  if(i < m_observers.size())
    m_observers.erase(m_observers.begin() + i);
  
}

/**
@brief Calls OnUpdate with the input notification code on all observers
@params code_notify - The notification code to be sent
*/
void Subject::Notify (int code_notify) {
  
  /// Update all observers
  for (int i = 0; i < m_observers.size(); i++)
    (m_observers[i])->OnUpdate(this, code_notify); 
  
}

/**
@brief Wrapper function for CreateNewProject
@params filename - name of the rootfly file
@params directory - The image directory
*/
void Controller::C_CreateNewProject(const CString& filename, const CString& directory)
{
  /// Get information from imagefileformat to name and create various data structures
  const ImageFileFormat& iff = m_data.GetExperiment().GetImageFileFormat();
  ImageInfo temp_image_info = iff.GetImageInfoOfFirstFile(directory);
  
  C_SetFilename(filename);
  C_SetModify(false);
  C_SetExperiment(temp_image_info.expt);
  C_SetTube(temp_image_info.tube);
  C_SetWindowNoRemap(temp_image_info.window);	//changed to not break. 
  C_SetSessionNoRemap(temp_image_info.session);
  
  Notify(ON_INIT);
}



/**
@brief Wrapper function for LoadFile
@param fname - the name of the file to be loaded
@returns 0 for success, nonzero for failure (see the ERRORS enum in DataStructures.h)
*/
int Controller::C_LoadFile(const CString& fname)
{
  int temp_int; 
  Experiment temp_exp;
  
  temp_exp.SetOldFile(false);
  Settings& settings = m_data.GetSettings();
  
  /// Load file and set the filename in the data structures
  temp_int = temp_exp.Load((char *) (LPCTSTR)fname, &settings);
  if (temp_int == ERRORS::RFY_LOADERROR_SUCCESS)
  {
    m_data.GetExperiment().ClearAll();  // !!! This should not be needed; but without it we get triangles displayed connecting first and last points of root -- JCL 4/14/10
    m_data.GetExperiment() = temp_exp; 
    C_SetFilename(fname);
    
    /// Set status saved
    C_SetModify(false);
    Notify(NEW_FILE_LOADED);
  }
  
  return temp_int;
}

/**
@brief Wrapper function for SaveFile
@param fname - Name of the file to be saved
@returns 0 for sucess, -1 for failure
*/
int Controller::C_SaveFile(const CString& fname, bool clear_modify_flag)
{
  int temp_int;
  g_debug_logfile.Print("Controller::C_SaveFile 1\n"); 
  const Experiment& exp = m_data.GetExperiment();
  g_debug_logfile.Print("Controller::C_SaveFile 2 %x\n", exp); 
  const Settings& settings = m_data.GetSettings();
  g_debug_logfile.Print("Controller::C_SaveFile 3 %x\n", settings); 
  
  /// Save the file in the experiment
  temp_int = exp.Save((char *) (LPCTSTR)fname, &settings);
  if (temp_int >= 0)
  {
    g_debug_logfile.Print("Controller::C_SaveFile 4\n"); 
    Notify(FILE_SAVED);
    g_debug_logfile.Print("Controller::C_SaveFile 5\n"); 
    /// Set status saved
    if (clear_modify_flag)  C_SetModify(false);
    g_debug_logfile.Print("Controller::C_SaveFile 6\n"); 
  }
  else
  {
    assert( temp_int == -1 );
    g_debug_logfile.Print("Controller::C_SaveFile error\n"); 
    CString str;
    str.Format("Unknown error saving file '%s'", (const char*) fname);
    AfxMessageBox(str);
  }
  
  return temp_int;
}

/**
@brief Returns RootInstance from root_id based on current ApplicationData
@param root_id Identification number for the root
@return Pointer to RootInstance
*/
RootInstance *Controller::C_GetRootInstance(int root_id)
{
  /// Get the current tube and session number from the ApplicationData
  ApplicationData& app_data = m_data.GetApplicationData();
  CString tube_tag = app_data.GetCurrentTubeName();
  int session = app_data.GetCurrentSession();
  
  /// Get the tube from the experiment using the tag from the ApplicationData
  Experiment& exp = m_data.GetExperiment();
  Tube* tube = exp.GetTubeByName(tube_tag);
  /// Failure if tube is null
  if(tube == NULL)
  {
    return NULL;
  }
  
  /// Get the root from the tube by the Root ID
  Root* temp_root = tube->GetRoot(root_id);
  /// Failure if root is null
  if(temp_root == NULL)
  {
    return NULL;
  }
  
  /// Get the root instance using the session number from ApplicationData
  RootInstance* root_instance = temp_root->GetRootInstance(session);
  
  return root_instance;
}

/**
@brief Returns RootInstance from root_id based on current ApplicationData
@param root_id Identification number for the root
@return Pointer to RootInstance
*/
RootInstance *Controller::C_GetRootInstance(int root_id, int session)
{
  /// Get the current tube and session number from the ApplicationData
  ApplicationData& app_data = m_data.GetApplicationData();
  CString tube_tag = app_data.GetCurrentTubeName();  
  
  /// Get the tube from the experiment using the tag from the ApplicationData
  Experiment& exp = m_data.GetExperiment();
  Tube* tube = exp.GetTubeByName(tube_tag);
  /// Failure if tube is null
  if(tube == NULL)
  {
    return NULL;
  }
  
  /// Get the root from the tube by the Root ID
  Root* temp_root = tube->GetRoot(root_id);
  /// Failure if root is null
  if(temp_root == NULL)
  {
    return NULL;
  }
  
  /// Get the root instance using the session number from ApplicationData
  RootInstance* root_instance = temp_root->GetRootInstance(session);
  
  return root_instance;
}

/**
@brief Returns RootInstance from root_id based on current ApplicationData
@param root_id Identification number for the root
@return Pointer to RootInstance
*/
RootInstance *Controller::C_CreateRootInstance(int root_id, int session)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  CString tube_tag = app_data.GetCurrentTubeName();
  
  Experiment& exp = m_data.GetExperiment();
  Tube* tube = exp.GetTubeByName(tube_tag);
  /// Failure if tube is null
  if(tube == NULL)
  {
    return NULL;
  }
  
  Root* temp_root = tube->GetRoot(root_id);
  /// Failure if root is null
  if(temp_root == NULL)
  {
    return NULL;
  }
  
  RootInstance* root_instance = temp_root->CreateRootInstance(session);
  
  /// Set status unsaved
  C_SetModify(true);
  Notify(ROOT_INSTANCE_ADDED);
  return root_instance;
}

/**
@brief Wrapper function for CreatePoint
@params root_id - index of root instance to which the point is created
@returns Pointer to RPoint created
*/
RPoint* Controller::C_CreatePoint(int root_id, int session)
{
  RPoint *temp_point;
  RootInstance *root_inst = C_GetRootInstance(root_id, session);
  
  //bool executed = Controller::DoCommand(new UC_CreatePoint(root_inst,root_id));
  /// Failure if root instance is null
  if(root_inst == NULL)
  {
    return NULL;
  }
  
  temp_point = root_inst->CreatePoint();
  Notify(POINT_ADDED);
  /// Set status unsaved
  C_SetModify(true);
  return temp_point;
}

/**
@brief Wrapper function for CreateMidPoint
@params root_id - index of root index to which the point is created
@params index - index of point on root that the new point is going to be
inserted after
@returns Pointer to RPoint created
*/
RPoint* Controller::C_CreateMidPoint(int root_id, int session, int index)
{
  bool executed = Controller::DoCommand(new UC_AddPoint(&m_data, root_id, session, index));
  
  // Since DoCommand returns a bool and CreateMidPoint expects a RPoint to be returned,
  // use the index of the newly created point to get a pointer to it. 
  RPoint *mid_point;
  RootInstance *root_inst = C_GetRootInstance(root_id, session);
  mid_point = root_inst->GetPoint(index+1);
  
  
  Notify(POINT_ADDED);
  
  /// Set status unsaved
  C_SetModify(true);
  
  return mid_point;
}


/**
@brief Wrapper function for DeletePoint
@params index - index of point to be deleted
@params root_id - index of root instance from which point will be deleted
@returns true for sucess, false for failure
*/
bool Controller::C_DeletePoint(int index, int root_id, int session)
{
  bool executed = Controller::DoCommand(new UC_DeletePoint(&m_data,root_id,index, session));
  Notify(POINT_REMOVED);
  /// Set status unsaved
  C_SetModify(true);
  
  return executed;
}

bool Controller::C_ExtendEndpoint(RootInstance &old_ri, int root_id, int session)
{
  DoCommand(new UC_ExtendEndpoint(&m_data, old_ri, root_id, session));
  return true;
}

/**
@brief Function to move all roots in current Tube Window Session
@params window - window to move points in
@params offset_x - x coordinate offset of moved point
@params offset_y - y coordinate offset of moved point
@returns true for sucess, false for failure
*/
bool Controller::C_MoveRootsInWindow(int window, int session ,int offset_x, int offset_y){
  
  
  //get all roots in tube/window/session
  Experiment& exp = m_data.GetExperiment();
  Tube *tube = exp.GetTubeByName(m_data.GetApplicationData().GetCurrentTubeName());
  /// Failure if tube is null
  if( tube == NULL ) return false;
  
  /// Get all roots
  vector<Root> *roots = tube->GetRootsInWindow(window);
  if (roots)
  {
    vector<Root>::iterator iter;
    //for each root
    for(iter = roots->begin();iter != roots->end();iter++)
      C_MoveRootSess(iter->GetID(), session, offset_x, offset_y);
  }
  
  
  return true;
}

// The Undo version of move all roots in window. This is called when the user has completed the 
// action of moving roots around in a window.
bool Controller::C_MoveRootsInWindow_U(int window, int session, int offset_x, int offset_y, int total_offset_x, int total_offset_y)
{
  
  bool executed = Controller::DoCommand(new UC_MoveRootsInWindow(&m_data,window, session, total_offset_x,total_offset_y));
  
  
  return true;
}

/**
@brief Function to move a root for a particular session
@params root_id - index of root to be moved
@params session - session in which to move the root
@params offset_x - x coordinate offset of moved point
@params offset_y - y coordinate offset of moved point
@returns true for sucess, false for failure
*/
bool Controller::C_MoveRootSess(int root_id, int session, int offset_x, int offset_y)
{
  
  Experiment& exp = m_data.GetExperiment();
  Tube *tube = exp.GetTubeByName(m_data.GetApplicationData().GetCurrentTubeName());
  /// Failure if tube is null
  if( tube == NULL ) return false;
  
  Root *r_root = tube->GetRoot(root_id);
  /// Failure if root is null
  if( r_root == NULL ) return false;
  
  RootInstance *r_inst = r_root->GetRootInstance(session);
  /// Failure if root instance is null
  if( r_inst == NULL ) return false;
  
  /// For all points
  for(int i = 0;i<r_inst->GetAllPoints()->size();i++)
    C_MovePoint(i, root_id, session, r_inst->GetPoint(i)->x+offset_x,r_inst->GetPoint(i)->y+offset_y);//may need to update this to be a bit different with x and y
  
  /// Move Diameter (assumes only one diameter)

  C_MoveDiameter(0,root_id, session, r_inst->GetDiameter(0)->r_point.x + offset_x, r_inst->GetDiameter(0)->r_point.y + offset_y);

  return true;
}

// The Undo version of move a single root. Called when the user has finished moving
// a root in a window.
bool Controller::C_MoveRootSessU(int root_id, int session, int offset_x, int offset_y, int total_offset_x, int total_offset_y)
{
  bool executed = Controller::DoCommand(new UC_MoveRootSess(&m_data,root_id,session,total_offset_x,total_offset_y));
  
  return true;
}

/**
@brief Wrapper function for MovePoint
@param index - index of point to be moved
@params root_id - index of root to be moved
@params offset_x - x coordinate offset of moved point
@params offset_y - y coordinate offset of moved point
@returns true for sucess, false for failure
*/
bool Controller::C_MovePoint(int index, int root_id, int session, int x, int y)
{ 
  RootInstance *root_inst = C_GetRootInstance(root_id, session); 
	
  /// Failure if root instance is null
  if(root_inst == NULL)     return false;
  
  /// Move the point
  if(root_inst->MovePoint(index, x, y))
  {
    /// Set status unsaved
    C_SetModify(true);
    Notify(POINT_MOVED);
    return true;
  }
  else return false;
  
}

// the Undo version of Move Point
bool Controller::C_MovePoint_U(int index, int root_id, int session, int new_x, int new_y, RootInstance * prev)
{
	RootInstance *root_inst = C_GetRootInstance(root_id, session);
	RDiameter *new_diam = root_inst->GetDiameter(0);
	RDiameter *old_diam = prev->GetDiameter(0);
	RPoint *new_pt = root_inst->GetPoint(index);
	RPoint *old_pt = prev->GetPoint(index);

  bool executed = Controller::DoCommand(new UC_MovePoint(&m_data, index, root_id, session, *new_pt, *old_pt, *new_diam, *old_diam));
  
  return executed;
}




/**
@brief Wrapper function for MovePoint
@param index - index of point to be moved
@params root_id - index of root to be moved
@params offset_x - x coordinate offset of moved point
@params offset_y - y coordinate offset of moved point
@returns true for sucess, false for failure
*/
void Controller::C_MovePointSess(int index, int root_id, int session, int x, int y)
{ 
  RootInstance *root_inst = C_GetRootInstance(root_id, session);
  /// Failure if root instance is null
  if(root_inst == NULL)     return;
  
  /// Move the point
  root_inst->MovePoint(index, x, y);  
}

/**
@brief The is a wrapper for ChangeWindowOfPoint
@params root_id - index of root instance to which point belongs
@params index - index of point whose window is to be changed
@params w - window to move point to
@returns true for sucess, false for failure
*/
bool Controller::C_ChangeWindowOfPoint(int index, int root_id, int session, int w)
{
  RootInstance *root_inst = C_GetRootInstance(root_id, session);
  /// Failure if root instance is null
  if(root_inst == NULL)     return false;
  
  if(root_inst->ChangeWindowOfPoint(index, w))
  {
    /// Set status unsaved
    C_SetModify(true);    
    return true;
  }
  else return false;
}

/**
@brief Wrapper function for CreateDiameter
@params root_id - index of root instance to which diameter will be added
@returns Pointer to created RDiameter
*/
RDiameter* Controller::C_CreateDiameter(int root_id)
{
  RDiameter *temp_diam;
  RootInstance *root_inst = C_GetRootInstance(root_id);
  
  /// Failure if root instance is null
  if(root_inst == NULL)
  {
    return NULL;
  }
  
  temp_diam = root_inst->CreateDiameter();
  Notify(DIAMETER_ADDED);
  /// Set status unsaved
  C_SetModify(true);
  return temp_diam;
}

/**
@brief Wrapper function for DeleteDiameter
@params root_id - index of root to which diameter belongs
@params index - index of diameter to be deleted
@returns true for sucess, false for failure
*/
bool Controller::C_DeleteDiameter(int index, int root_id, int session)
{
  bool temp_bool;
  RootInstance *root_inst = C_GetRootInstance(root_id, session);
  
  /// Failure if root instance is null
  if(root_inst == NULL) return false;
  
  
  temp_bool = root_inst->DeleteDiameter(index);
  Notify(DIAMETER_REMOVED);
  /// Set status unsaved
  C_SetModify(true);
  return temp_bool;
}

/**
@brief Wrapper function for MoveDiameter
@params index - index of diameter to be moved
@params root_id - index of root to be moved
@params x - x coordinate offset of moved diameter
@params y - y coordinate offset of moved diameter
@returns true for sucess, false for failure
*/
bool Controller::C_MoveDiameter(int index, int root_id, int session, int x, int y)
{  
  RootInstance *root_inst = C_GetRootInstance(root_id, session);
  
  /// Failure if root instance is null
  if(root_inst == NULL)     return false;
  
  if(root_inst->MoveDiameter(index, x, y))
  {
    /// Set status unsaved
    C_SetModify(true);
    Notify(DIAMETER_MOVED);
    return true;
  }
  else return false;
}

bool Controller::C_MoveDiameter(int index, int root_id, int session, const RPoint& r)
{
  bool success1, success2;
  success1 = C_MoveDiameter(index, root_id, session, r.x, r.y); 
  success2 = C_ChangeWindowOfDiameter(index, root_id, session, r.window);
  return success1 && success2;
}

// the Undo version of move diameter. Called when the user has finished
// moving the diameter to a different location on a root.
bool Controller::C_MoveDiameter_U(int index, int root_id, int session, const RPoint& old_r, const RPoint& new_r)
{
  bool executed = Controller::DoCommand(new UC_MoveDiameter(&m_data,index, root_id, session, old_r,new_r));
  
  return true;
}

/**
@brief Wrapper function for MoveDiameter
@params index - index of diameter to be moved
@params root_id - index of root to be moved
@params x - x coordinate offset of moved diameter
@params y - y coordinate offset of moved diameter
@returns true for sucess, false for failure
*/
bool Controller::C_MoveDiameterSess(int index, int root_id, int session, int x, int y)
{  
  RootInstance *root_inst = C_GetRootInstance(root_id, session);
  
  /// Failure if root instance is null
  if(root_inst == NULL)     return false;
  
  if(root_inst->MoveDiameter(index, x, y))
  {
    /// Set status unsaved
    C_SetModify(true);
    Notify(DIAMETER_MOVED);
    return true;
  }
  else return false;
}

/**
@brief The is a wrapper for ChangeWindowOfPoint
@params root_id - index of root instance to which point belongs
@params index - index of point whose window is to be changed
@params w - window to move point to
@returns true for sucess, false for failure
*/
bool Controller::C_ChangeWindowOfDiameter(int index, int root_id, int session, int w)
{
  RootInstance *root_inst = C_GetRootInstance(root_id, session);
  
  /// Failure if root instance is null
  if(root_inst == NULL)     return false;
  
  if(root_inst->ChangeWindowOfDiameter(index, w))
  {
    /// Set status unsaved
    C_SetModify(true);
    return true;
  }
  else return false;
}

/**
@brief Wrapper function for ResizeDiameter
@params index - index of diameter to be resized
@params root_id - index of root to be resized
@params radius - new radius of diameter
@returns true for sucess, false for failure
*/
bool Controller::C_ResizeDiameter_U(int index, int root_id, int session, double old_radius, double new_radius)
{  
  bool executed = Controller::DoCommand(new UC_ResizeDiameter(&m_data,index,root_id, session, old_radius,new_radius));
  
  /// Set status unsaved
  C_SetModify(true);
  Notify(STATUS_CHANGED);
  
  return executed;
  
}

bool Controller::C_ResizeDiameter(int index, int root_id, int session, double radius)
{

	RootInstance *root_inst = C_GetRootInstance(root_id, session);

  
  /// Failure if root instance is null
  if(root_inst == NULL)     return false;
  
  if(root_inst->ResizeDiameter(index, radius))
  {
    /// Set status unsaved
    C_SetModify(true);
    Notify(DIAMETER_CHANGED);
    return true;
  }
  else return false;
}

/**
@brief Wrapper function for ResizeDiameter
@params index - index of diameter to be resized
@params root_id - index of root to be resized
@params radius - new radius of diameter
@returns true for sucess, false for failure
*/
bool Controller::C_ResizeDiameterSess(int index, int root_id, int session, double radius)
{  
  RootInstance *root_inst = C_GetRootInstance(root_id, session);
  
  /// Failure if root instance is null
  if(root_inst == NULL)     return false;
  
  if(root_inst->ResizeDiameter(index, radius))
  {
    /// Set status unsaved
    C_SetModify(true);
    Notify(DIAMETER_CHANGED);
    return true;
  }
  else return false;
}

/**
@brief Wrapper function for CreateRoot
@params primary_window - window in which to create root
@returns Pointer to new root
*/
// Note:  Every time we call C_CreateRoot, we immediately call C_CreateRootInstance.
// As a result, this function may become obsolete and replaced with C_CreateRoot below
// that also takes a RootInstance. -- STB
Root* Controller::C_CreateRoot(int primary_window)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  CString tube_tag = app_data.GetCurrentTubeName();  
  Experiment& exp = m_data.GetExperiment();
  
  Tube* tube = exp.GetTubeByName(tube_tag);
  
  /// If tube does not exist, create a new tube
  if (tube == NULL)  tube = exp.CreateTube(tube_tag);
  assert(tube);
  
  Root* temp_root = tube->CreateRoot(primary_window);
  
  Notify(ROOT_ADDED);
  /// Set status unsaved
  C_SetModify(true);
  return temp_root;
}

// Needs to be tested before being used -- STB
void Controller::C_CreateRoot(int primary_window, const RootInstance& ri, CString default_cond)
{
  
  DoCommand(new UC_CreateRoot(&m_data, primary_window, ri, default_cond));
  
  Notify(ROOT_ADDED);
  /// Set status unsaved
  C_SetModify(true);
}

/**
This function copies roots from one session to another.
Either copies only the current tube and window, or all tubes and windows,
depending upon whether 'curr_window' is nonnegative.

  @param from Source session to obtain root information.
  @param to Destination session to receive root information from the specified source.
  @param tubename The tag of the current working tube (ignored if curr_window is -1)
  @param curr_window Specifies the window to perform the copy within, or -1 to copy 
  roots of all windows and all tubes.
  @param which_root  Which root to copy.  -1 means all roots in current window.
  This parameter is ignored if curr_window == -1
*/

void Controller::C_CopySession(int from, int to, const CString &tubename, int curr_window, int which_root) 
{
  bool executed = DoCommand(new UC_CopySession(&m_data, from, to, tubename, curr_window, which_root));
  /// Set status unsaved
  C_SetModify(true);
  Notify(COPY_PREVIOUS);
}


/**
@brief Wrapper function for CreateTube
@params tag - name of tube to be created
@returns Pointer to new tube
*/
Tube* Controller::C_CreateTube(const CString& tag)
{
  Tube *temp_tube;
  
  Experiment& exp = m_data.GetExperiment();
  temp_tube = exp.CreateTube(tag);
  /// Set so that a root can be added
  assert( temp_tube->GetRootIDCounter() == 1 );
  
  Notify(TUBE_ADDED);
  /// Set status unsaved
  C_SetModify(true);
  return temp_tube;
}

/**
@brief Wrapper function for DeleteRoot
@params root_id - Index of root to be deleted
@returns true if sucess, false if failure
*/
bool Controller::C_DeleteRoot(int root_id)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  CString tube_tag = app_data.GetCurrentTubeName();
  Experiment& exp = m_data.GetExperiment();
  Tube* tube = exp.GetTubeByName(tube_tag);
  
  bool executed = DoCommand(new UC_DeleteRoot(&m_data, tube_tag, root_id, tube->GetRoot(root_id)));
  
  
  Notify(ROOT_DELETED);
  /// Set status unsaved
  C_SetModify(true);
  
  return executed;
}

/**
@brief Wrapper function for SetCurrentTubeName
@params tube - tube to be set as current tube
*/
void Controller::C_SetTube(const CString& tube)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  Experiment& exp = m_data.GetExperiment();
  
  app_data.SetCurrentTubeName(tube);
  /// Reload the image buffers for the new tube
  app_data.ReloadImageBuffers(exp.GetImageDir(),exp.GetImageFileFormat(),m_data.GetSessionCollapser());
  Notify(TUBE_CHANGED);
}

/**
@brief Wrapper function for SetCurrentWindow
@params window - int to change current window to
*/
void Controller::C_SetWindow(int window)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  Experiment& exp = m_data.GetExperiment();
  SessionCollapser& sess_coll = m_data.GetSessionCollapser();
  int changed = window - app_data.GetCurrentWindow();
  
  app_data.SetCurrentWindow(window);
  sess_coll.RemapSessionIndices();
  
  app_data.ReloadImageBuffers( exp.GetImageDir(), exp.GetImageFileFormat(), m_data.GetSessionCollapser() );
  Notify(WINDOW_CHANGED);
}

//same as above without remapping. Remapping crashes the program if done before the SetNavigation call in RootflyDlg::OnFileNew.
void Controller::C_SetWindowNoRemap(int window)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  Experiment& exp = m_data.GetExperiment();
  int changed = window - app_data.GetCurrentWindow();
  
  app_data.SetCurrentWindow(window);
  
  app_data.ReloadImageBuffers( exp.GetImageDir(), exp.GetImageFileFormat(), m_data.GetSessionCollapser() );
  Notify(WINDOW_CHANGED);
}

/**
@brief Wrapper function for SetCurrentSession
@params session - int to change current session to
*/
void Controller::C_SetSession(int session)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  Experiment& exp = m_data.GetExperiment();
  SessionCollapser& sess_coll = m_data.GetSessionCollapser();
  
  int changed = session-app_data.GetCurrentSession();
  
  app_data.SetCurrentSession(session);
  sess_coll.RemapSessionIndices();
  
  app_data.ReloadImageBuffers( exp.GetImageDir(), exp.GetImageFileFormat(), m_data.GetSessionCollapser() );
  Notify(SESSION_CHANGED);
}

//same as above without remapping
void Controller::C_SetSessionNoRemap(int session)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  Experiment& exp = m_data.GetExperiment();
  
  int changed = session-app_data.GetCurrentSession();
  
  app_data.SetCurrentSession(session);
  
  app_data.ReloadImageBuffers( exp.GetImageDir(), exp.GetImageFileFormat(), m_data.GetSessionCollapser() );
  Notify(SESSION_CHANGED);
}

/**
@brief Wrapper function for SetShowRootID
@params status - Updated status of Show Root IDs checkbox
*/
void Controller::C_SetShowRootID(bool status)
{
  Settings& settings = m_data.GetSettings();
  
  settings.SetShowRootID(status);
  Notify(SHOW_ROOT_ID_CHANGED);
}

/**
@brief Wrapper function for SetShowDeadRoots
@params status - Updated status of Show Dead Roots checkbox
*/
void Controller::C_SetShowDeadRoots(bool status)
{
  Settings& settings = m_data.GetSettings();
  
  settings.SetShowDeadRoots(status);
  Notify(SHOW_DEAD_ROOTS_CHANGED);
}

/**
@brief Wrapper function for SetShowRoots
@params status - Updated status of Show Roots checkbox
*/
void Controller::C_SetShowRoots(bool status)
{
  Settings& settings = m_data.GetSettings();
  
  settings.SetShowRoots(status);
  Notify(SHOW_ROOTS_CHANGED);
}


/**
@brief Wrapper function for SetShowDiameters
@params status - Updated status of Show Diameters checkbox
*/
void Controller::C_SetShowDiameters(bool status)
{
  Settings& settings = m_data.GetSettings();
  
  settings.SetShowDiameters(status);
  Notify(SHOW_DIAMETERS_CHANGED);
}

/**
@brief Wrapper function for SetExperiment
@param experiement - name of new experiment to be set
*/
void Controller::C_SetExperiment(const CString& experiment)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  
  app_data.SetExperiment(experiment);
  Notify(EXPERIMENT_CHANGED);
}

/**
@brief Wrapper function for SetCurrentImage
@params current_image - image to make the current image
*/
void Controller::C_SetCurrentImage(const CString& current_image)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  Experiment& exp = m_data.GetExperiment();
  
  app_data.SetCurrentImage(current_image);
  /// Reload the image buffer for the new image
  app_data.ReloadImageBuffers(exp.GetImageDir(),exp.GetImageFileFormat(), m_data.GetSessionCollapser());
  Notify(CURRENT_IMAGE_CHANGED);
}

/**
@brief Wrapper function for SetSessionDate
@params date - New session date
*/
void Controller::C_SetSessionDate(const CString& date)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  
  app_data.SetSessionDate(date);
  Notify(SESSION_DATE_CHANGED);
}

/**
@brief Wrapper function for SetOperator
@params operator_initials - New operator initials
*/
void Controller::C_SetOperator(const CString& operator_initials)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  
  app_data.SetOperator(operator_initials);
  Notify(OPERATOR_CHANGED);
}

/**
@brief Sets the save status of the file.  Notifies if the file has gone from saved to unsaved
or from unsaved to saved.  This allows the GUI to know where the file needs to be
or has been saved.
@params modify - true if the file has been changed, false if the file has been saved
*/
void Controller::C_SetModify(bool modify)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  
  /// Get the current save status
  bool temp_status = app_data.GetModify();
  
  /// If the status has changed, update the status in the data scructures and notify
  if(temp_status != modify)
  {
    app_data.SetModify(modify);
    Notify(modify ? SAVE_STATUS_UNSAVED : SAVE_STATUS_SAVED);
  }
}

/**
@brief Wrapper function for SetFilename
@params filename - new filename
*/
void Controller::C_SetFilename(const CString& filename)
{
  ApplicationData& app_data = m_data.GetApplicationData();
  
  app_data.SetFilename(filename);
  /// Fixes an error in which the save status of a file whose name has been changed
  /// will not update save status until changed
//  C_SetModify(true);
  Notify(FILENAME_CHANGED);
}

/**
@brief Wrapper function for SetImageDir
@params imgdir - new directory to pull images from
*/
void Controller::C_SetImageDir(const CString& imgdir)
{
  Experiment& exp = m_data.GetExperiment();
  exp.SetImageDir(imgdir);
  Notify(IMAGEDIR_CHANGED);
}

/**
@brief Wrapper function for SetImageFileFormat
@params iff - the new imagefileformat
*/
void Controller::C_SetImageFileFormat(ImageFileFormat iff)
{
  Experiment& exp = m_data.GetExperiment();
  exp.SetImageFileFormat(iff);
  Notify(IMAGEFILEFORMAT_CHANGED);
}

// obsolete !!!
/**
@brief Wrapper function for GetImageFileFormat - returns the imagefileformat
@return The Imagefileformat
*/
//ImageFileFormat Controller::C_GetImageFileFormat() const
//{
//  const Experiment& exp = m_data.GetExperiment();
//  return exp.GetImageFileFormat();
//}

/**
@brief Calls ClearAll for the root instance in the current TWS
@params root_id - ID of rootinstance to clear
@return - true if sucessful, false if unsucessful
*/
bool Controller::C_ClearAllRootInstance(int root_id)
{
  bool temp_bool;
  RootInstance *root_instances = C_GetRootInstance(root_id);
  /// Failure if root instance is null
  if(root_instances == NULL) {
    return false;
  }
  
  temp_bool = root_instances->ClearAll();
  Notify(ROOT_INSTANCE_CLEAR);
  /// Set status unsaved
  Controller::C_SetModify(true);
  return temp_bool;
}

/**
@brief Calls ClearAll for the root in the current TWS
@params root_id - index of root which will be cleared
@return - true if sucessful, false if unsucessful
*/
bool Controller::C_ClearAllRoot(int root_id)
{
  bool temp_bool;
  
  /// Get to the root
  ApplicationData& app_data = m_data.GetApplicationData();
  CString tube_tag = app_data.GetCurrentTubeName();
  Experiment& exp = m_data.GetExperiment();
  Tube* tube = exp.GetTubeByName(tube_tag);
  /// Failure if the tube is null
  if(tube == NULL) {
    return false;
  }
  
  Root* temp_root = tube->GetRoot(root_id);
  /// Failure if the root is null
  if(temp_root == NULL) {
    return false;
  }
  
  temp_bool = temp_root->ClearAll();
  Notify(ROOT_CLEAR);
  /// Set status unsaved
  Controller::C_SetModify(true);
  return temp_bool;
  
}

/**
@brief Calls ClearAll for the current tube (clears everything in the current tube)
@return - True if sucessful, false if unsucessful
*/
bool Controller::C_ClearAllTube()
{
  bool temp_bool;
  
  /// Get to the current tube
  ApplicationData& app_data = m_data.GetApplicationData();
  CString tube_tag = app_data.GetCurrentTubeName();
  Experiment& exp = m_data.GetExperiment();
  Tube* tube = exp.GetTubeByName(tube_tag);
  /// Failure if tube is null
  if(tube == NULL) {
    return false;
  }
  
  temp_bool = tube->ClearAll();
  Notify(TUBE_CLEAR);
  /// Set status unsaved
  Controller::C_SetModify(true);
  return temp_bool;
}

/**
@brief Calls ClearAll for the experiment (clears all information in the current experiment)
@returns true if sucessful, false if unsucessful
*/
bool Controller::C_ClearAllExperiment()
{
  bool temp_bool;
  
  /// Get the Experiment
  Experiment& exp = m_data.GetExperiment();
  
  temp_bool = exp.ClearAll();
  Notify(EXPERIMENT_CLEAR);
  /// Set status unsaved
  return temp_bool;
}

/**
@brief Wrapper function to read settings from registry
*/
void Controller::C_ReadUserPreferences()
{
  /// Get the user preferences
  Settings& settings = m_data.GetSettings();
  
  settings.ReadUserPreferences();
  Notify(PREFERENCES_READ);
}

/**
@brief Wrapper function to save settings to registry
*/
void Controller::C_SaveUserPreferences()
{
  /// Get the user preferences
  const Settings& settings = m_data.GetSettings();
  
  settings.SaveUserPreferences();
  Notify(PREFERENCES_SAVED);
}

/**
@brief Wrapper function to set min number of windows
@params windows - the new min number of windows
*/
void Controller::C_SetMinWindows(int windows)
{
  Experiment& exp = m_data.GetExperiment();
  
  exp.SetMinWindows(windows);
  Notify(MIN_WINDOWS_CHANGED);
}

/**
@brief Wrapper function to set min number of sessions
@params sessions - the new min number of sessions
*/
void Controller::C_SetMinSessions(int sessions)
{
  Experiment& exp = m_data.GetExperiment();
  
  exp.SetMinSessions(sessions);
  Notify(MIN_SESSIONS_CHANGED);
}

/**
@brief Wrapper function to set max number of windows
@params windows - the new max number of windows
*/
void Controller::C_SetMaxWindows(int windows)
{
  Experiment& exp = m_data.GetExperiment();
  
  exp.SetMaxWindows(windows);
  Notify(MAX_WINDOWS_CHANGED);
}

/**
@brief Wrapper function to set max number of sessions
@params sessions - the new max number of sessions
*/
void Controller::C_SetMaxSessions(int sessions)
{
  Experiment& exp = m_data.GetExperiment();
  
  exp.SetMaxSessions(sessions);
  Notify(MAX_SESSIONS_CHANGED);
}

/**
@brief Wrapper function to set mm conversion
@params conversion - the new mm-to-pixel conversion
*/
void Controller::C_SetMMConversion(float conversion)
{
  Experiment& exp = m_data.GetExperiment();
  
  exp.SetMMConversion(conversion);
  Controller::C_SetModify(true);
  Notify(MM_CONVERSION_CHANGED);
}

/**
@brief Wrapper function to set root drawing color
@params color - the new color for roots being drawn
*/
void Controller::C_SetRootDrawingColor(COLORREF color)
{
  Settings& settings = m_data.GetSettings();
  
  settings.SetRootDrawingColorLive(color);
  Notify(ROOT_DRAWING_COLOR_CHANGED);
}

/**
@brief Wrapper function to set selected root drawing color
@params color - the new color for the selected root
*/
void Controller::C_SetRootDrawingColorSelected(COLORREF color)
{
  Settings& settings = m_data.GetSettings();
  
  settings.SetRootDrawingColorLiveSelected(color);
  Notify(SELECTED_ROOT_DRAWING_COLOR_CHANGED);
}

/**
@brief Wrapper function to set dead root drawing color
@params color - the new color for dead roots
*/
void Controller::C_SetRootDrawingColorDead(COLORREF color)
{
  Settings& settings = m_data.GetSettings();
  
  settings.SetRootDrawingColorDead(color);
  Notify(SELECTED_ROOT_DRAWING_COLOR_CHANGED);
}

/**
@brief Wrapper function to set dead selected root drawing color
@params color - the new color for dead roots
*/
void Controller::C_SetRootDrawingColorDeadSelected(COLORREF color)
{
  Settings& settings = m_data.GetSettings();
  
  settings.SetRootDrawingColorDeadSelected(color);
  Notify(DEAD_SELECTED_ROOT_DRAWING_COLOR_CHANGED);
}

/**
*/
void Controller::C_SetNewRootColor(RootInstance::Color color)
{
  Settings& settings = m_data.GetSettings();
  
  settings.SetNewRootColor(color);
}

/**
*/
void Controller::C_SetNewRootIsAlive(bool ia)
{
  Settings& settings = m_data.GetSettings();
  
  settings.SetNewRootIsAlive(ia);
}

/**
@brief Wrapper function to set root color
@params root_id - the index of the root whose color is to be changed
@params color_choise - root color to be set
*/
void Controller::C_SetRootColor(int root_id, int session, RootInstance::Color color_choice)
{
  
  bool executed = DoCommand(new UC_SetRootColor(&m_data,root_id, session, color_choice));
  
  /// Set status unsaved
  if(executed)
  {
    C_SetModify(true);
    Notify(COLOR_CHANGED);
  }
  
}

/**
@brief Wrapper function to set selected root status
@params root_id - index of root whose life status is to be changed
@params alive - true if the root is alive, false if the root is dead
*/
void Controller::C_SetRootAlive(int root_id, int session, bool alive)
{
  
  // Call DoCommand which executes the function and also
  // places the command at the back of the undo vector.
  bool executed = DoCommand(new UC_SetRootAlive(&m_data,root_id, session, alive));
  
  /// Set status unsaved
  if(executed)
  {
    C_SetModify(true);
    Notify(STATUS_CHANGED);
  }
  
}

/**
@brief Wrapper function to set selected root user defined code
@params root_id - index of root whos user condition code is to be changed
@params condition - user condition code to be attributed to root
*/
void Controller::C_SetConditionCode(int root_id, int session, CString condition)
{
  ApplicationData& app = m_data.GetApplicationData();
  Experiment& exp = m_data.GetExperiment();
  Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
  /// Failure if tube is null
  if(tube == NULL) return;
  Root *root = tube->GetRoot( root_id );
  /// Failure if root is null
  if(root == NULL) return;
  RootInstance *inst = root->GetRootInstance(session);
  /// Failure if root instance is null
  if(inst == NULL) return;
  CString prev = inst->GetConditionCode();
  /// If the condition code has changed
  if(prev != condition)
  {
    inst->SetConditionCode(condition);
    /// Set status unsaved
  }
  
  Notify(CONDITION_CODE_SELECTED);
}

/**
@brief Wrapper function to set selected root user defined code and make undoable
@params root_id - index of root whos user condition code is to be changed
@params condition - user condition code to be attributed to root
@params undo - wheater or not to make the function undoable
*/
void Controller::C_SetConditionCode(int root_id, int session, CString condition, bool undo)
{
  
  if(undo)
  {
    DoCommand(new UC_SetConditionCode(&m_data, root_id, session, condition));
    Notify(CONDITION_CODE_SELECTED);
  }
  else
  {
    ApplicationData& app = m_data.GetApplicationData();
    Experiment& exp = m_data.GetExperiment();
    Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
    /// Failure if tube is null
    if(tube == NULL) return;
    Root *root = tube->GetRoot( root_id );
    /// Failure if root is null
    if(root == NULL) return;
    RootInstance *inst = root->GetRootInstance(session);
    /// Failure if root instance is null
    if(inst == NULL) return;
    CString prev = inst->GetConditionCode();
    /// If the condition code has changed
    if(prev != condition)
    {
      inst->SetConditionCode(condition);
    }
  }
}

/**
@brief Wrapper function to add user defined code to list
@params condition - user condition code to be added
*/
void Controller::C_AddConditionCode(CString condition)
{
  Settings& settings = m_data.GetSettings();
  settings.AddConditionCode(condition);
  Controller::C_SetModify(true);
  Notify(CONDITION_CODE_ADDED);
}

/**
@brief Wrapper function to add user defined code to list
@params vector - Vector of condition codes to make the user condition codes
*/
void Controller::C_SetConditionSet(std::vector<CString> vector)
{
  Settings& settings = m_data.GetSettings();
  settings.AddConditionSet(vector);
  Notify(CONDITION_SET_CHANGED);
}

void Controller::C_SetConditionDefaultIndex(int index)
{
  Settings& settings = m_data.GetSettings();
  settings.SetConditionDefaultIndex(index);
  Notify(CONDITION_DEFAULT_INDEX_CHANGED);
}

/**
@brief Wrapper function to set the session dates for the project
@params dates - Vector of dates specified by the user to represent the sessions
*/

void Controller::C_SetSessionDates(std::vector<Date> dates)
{
  Experiment& exp = m_data.GetExperiment();
  exp.SetSessionDates(dates);
  Notify(SESSION_DATES_SET);
}
/**
@brief Wrapper function to set current image to a false positive
@params status - bool to indicate whether or not it should be a false positive
*/
void Controller::C_SetImageScanned(bool status)
{
  ApplicationData& app = m_data.GetApplicationData();
  Experiment& exp = m_data.GetExperiment();
  Tube *tube = exp.GetTubeByName(app.GetCurrentTubeName());
  /// Failure if tube is null
  if(tube == NULL) return;
  int curr_window = app.GetCurrentWindow();
  int curr_session = app.GetCurrentSession();  
  if(tube->GetImageScanned(curr_window,curr_session) != status)
  {
    tube->SetImageScanned(status, curr_window, curr_session);
    C_SetModify(true);
    Notify(IMAGED_SCANNED);
  }
}

void Controller::C_SetImageAnnotated_U(bool annotated, int window, int session)
{
  if( Controller::DoCommand(new UC_SetImageAnnotated(&m_data, annotated, window, session)) )
  {
    C_SetModify(true);
    Notify(IMAGE_ANNOTATED);
  }
}


void Controller::C_SetImageExists(bool status, CString tube_name, int window, int session)
{
  Experiment& exp = m_data.GetExperiment();
  Tube *tube = exp.GetTubeByName(tube_name);
  if (!tube)  tube = C_CreateTube(tube_name);
  assert(tube);
  if (!tube)  return;
  if(tube->GetImageExists(window,session) != status)
  {
    tube->SetImageExists(status, window, session);
  }
}

// rewrote this function to fix bug 2896256
int Controller::CalculateDaySince1900(const Date& date)
{
  int days_in_month[12] = {31,28,31,30,31,30,31,31,30,31,30,31};
  int i;
  int num_days;
  
  if (date.year < 1900)  return -1;
  
  //The 'month' variables stored in rootfly files are 1-indexed, so the month for loop begins
  //at i=1 to compensate.
  
  num_days = (date.year - 1900) * 365;
  for (i=1 ; i<date.month ; i++)  num_days += days_in_month[i-1];
  num_days += date.day;
  
  //This code handles leap years - JCL
  for(i = 1904; i<date.year; i = i + 4) num_days++;  // previous years (note:  1900 is not a leap year b/c divisible by 100 but not divisible by 400)
  if(date.month > 2 && date.year % 4 == 0) num_days++;  // if we are past Feb in current year  (will break in 2100)
  
  return num_days;
}

int Controller::CalculateDaysBetweenDates(const Date& start, const Date& end)
{
  int day1 = CalculateDaySince1900( start );
  int day2 = CalculateDaySince1900( end );
  
  // start date is bigger than end date
  if (day1 > day2)  return -1;
  
  return day2 - day1;
}

/**
* Handles the export functions for the main application
* @param filename String representing the file to be written
* @return nothing
*/
void Controller::C_Export(CString filename)
{
  
#define DISPLAY_CONDITION_CODE
  
  g_debug_logfile.Print("Controller::C_Export %s\n", (const char*) filename);
  
  int i,j,k,l, max_sessions=0, first_dead_session=0, first_alive_session=0, lifespan=0;
  CString tubename, color, status;        //Status == Alive/Dead
  double length=0, avg_diam=0, max_length;
  float conversion;
  const std::vector<Tube> *tubes;
  const std::vector<Root> *roots;
  const std::vector<RootInstance> *instances;
  vector<int> window_numbers;
  Date start, end;
  Experiment& exp = m_data.GetExperiment();;
  FILE *fp = fopen(filename, "wt");  
  //exp 
  tubes=exp.GetAllTubes();
  conversion = exp.GetMMConversion();
  
  if(fp == NULL) 
  {
    AfxMessageBox("Unable to open " + filename + " for writing");
    return;
  }
  
  if (tubes == NULL) 
  {
    return;
  }
  
  g_debug_logfile.Print("Controller::C_Export ntubes=%d\n", tubes->size());
  
  //Determine the maximum number of sessions in this experiment.
  for(i=0; i<tubes->size(); i++) 
  {
    g_debug_logfile.Print("Controller::C_Export tube %d\n", i);
    
    roots = tubes->at(i).GetAllRoots();
    g_debug_logfile.Print("Controller::C_Export nroots=%d\n", roots->size());
    
    for(j=0; j<roots->size(); j++) 
    {
      g_debug_logfile.Print("Controller::C_Export root %d\n", j);
      if(roots->at(j).GetLastSessionNumber() > max_sessions) 
      {
        max_sessions = roots->at(j).GetLastSessionNumber();
      }
    }
  }
  
  //Output columns
  g_debug_logfile.Print("Controller::C_Export Output columns, maxsessions=%d\n", max_sessions);
  fprintf(fp, "Tube,Window,RootID,Appeared (session number),Died (session number),Appeared,Died,Lifespan,Maximum Length,Maximum Diameter,");
  for(i=0; i<=max_sessions; i++) 
  {
    g_debug_logfile.Print("Controller::C_Export **0a i=%d \n", i);
    
#ifdef DISPLAY_CONDITION_CODE
    g_debug_logfile.Print("Controller::C_Export **0b i=%d \n", i);
    fprintf(fp, "Length date(%d),Diameter date(%d),Color date(%d),Condition Code date(%d),",i+1,i+1,i+1,i+1);
#else
    g_debug_logfile.Print("Controller::C_Export **0c i=%d \n", i);
    fprintf(fp, "Length date(%d),Diameter date(%d),Color date(%d)",i+1,i+1,i+1);
#endif
  }
  
  g_debug_logfile.Print("Controller::C_Export **0d\n");
  fprintf(fp, "\n");
  
  //Output the data, one root per line.
  g_debug_logfile.Print("Controller::C_Export Output data\n");
  for(i=0; i< tubes->size(); i++) 
  {
    g_debug_logfile.Print("Controller::C_Export tube=%d\n", i);
    roots = tubes->at(i).GetAllRoots();
    g_debug_logfile.Print("Controller::C_Export nroots=%d\n", roots->size());
    for(j=0; j< roots->size(); j++) 
    {      
      g_debug_logfile.Print("Controller::C_Export root=%d\n", j);
      first_dead_session = roots->at(j).GetFirstDeadSessionNumber();
      first_alive_session = roots->at(j).GetFirstSessionNumber();
      
      //If there are rootsessions available within this root....
      g_debug_logfile.Print("Controller::C_Export **1\n");
      if(first_alive_session != -1 && first_dead_session != -1) 
      {
        g_debug_logfile.Print("Controller::C_Export **2\n");
        
        //Output initial root information
        fprintf(fp, "%s,%d,%d,", tubes->at(i).GetTag(), roots->at(j).GetPrimaryWindow(), roots->at(j).GetDisplayableID());
        fprintf(fp, "%d,", first_alive_session); // -- STB 9/13/10
        
        //Make sure the values for the lifespan are the correct values.  These functions can return -1 and -2, which
        //  need to be handled to give this value the correct meaning.
        g_debug_logfile.Print("Controller::C_Export **3\n");
        if(first_dead_session == Root::ROOT_NEVER_DIED)
        {                              
          g_debug_logfile.Print("Controller::C_Export **4\n");
          start = exp.GetSessionDate(first_alive_session);
          end = exp.GetSessionDate(roots->at(j).GetLastSessionNumber());
          lifespan = CalculateDaysBetweenDates(start, end);
          fprintf(fp, "-,"); // 'first_dead_session' does not appear -- STB 9/13/10
          fprintf(fp, "%d/%d/%d,-,%d,", start.month, start.day, start.year, lifespan);
        }
        else
        {
          g_debug_logfile.Print("Controller::C_Export **5\n");
          start = exp.GetSessionDate(first_alive_session);
          end = exp.GetSessionDate(first_dead_session);
          lifespan = CalculateDaysBetweenDates(start, end);
          fprintf(fp, "%d,", first_dead_session); // -- STB 9/13/10
          fprintf(fp, "%d/%d/%d,%d/%d/%d,%d,", start.month, start.day, start.year, end.month, end.day, end.year, lifespan);
        }
        
        //Output the maximum length of the root and the maximum diameter on the root.
        g_debug_logfile.Print("Controller::C_Export **6\n");
        fprintf(fp, "%f,%f",  roots->at(j).GetMaxLength(conversion), roots->at(j).GetMaxDiameter(conversion));
        
        g_debug_logfile.Print("Controller::C_Export **7\n");
        //Start outputting where the root first exists (done by session number)
        for(k=1; k < roots->at(j).GetFirstSessionNumber(); k++)
        {
#ifdef DISPLAY_CONDITION_CODE
          fprintf(fp, ", , , ,");
#else
          fprintf(fp, ", , ,");
#endif
        }
        
        //Start outputting session specific information: color, length, diameter, and possibly the condition code
        instances = roots->at(j).GetAllRootInstances();
        
        //Pick the appropriate color to output to the file
        g_debug_logfile.Print("Controller::C_Export **8  ninstances=%d\n", instances->size());
        for(k=0; k < instances->size(); k++) 
        {
          if (!instances->at(k).IsAlive())  break;
          g_debug_logfile.Print("Controller::C_Export **9  inst=%d\n", k);
          max_length = 0.0;
          switch(instances->at(k).GetColor())
          {
          case 0:	  color = "White";       break;
          case 1:   color = "Light Brown"; break;
          case 2:   color = "Dark Brown";  break;
          case 3:   color = "Black";       break;
          default:  color = "White";       break;
          }
          instances->at(k).GetWindowNumbers(&window_numbers);	
          
          //Find the maximum length
          g_debug_logfile.Print("Controller::C_Export **10\n");
          for(l=0; l < window_numbers.size(); l++) 
          {
            max_length += instances->at(k).FindLengthInWindow(window_numbers.at(l), conversion);
          }
          
          //Output the root per-session information
#ifdef DISPLAY_CONDITION_CODE
          fprintf(fp, ",%f, %f,%s,%s", max_length, instances->at(k).GetAverageDiameter(conversion), color, instances->at(k).GetConditionCode());
#else
          fprintf(fp, ",%f, %f,%s", max_length, instances->at(k).GetAverageDiameter(conversion), color);
#endif
          g_debug_logfile.Print("Controller::C_Export **11\n");
        }
        fprintf(fp,"\n");
        g_debug_logfile.Print("Controller::C_Export **12\n");
      }
      
      g_debug_logfile.Print("Controller::C_Export **13\n");
    }
    g_debug_logfile.Print("Controller::C_Export **14\n");
  }
  g_debug_logfile.Print("Controller::C_Export **15\n");
  fclose(fp);
}


/**
@brief Wrapper function to set the image file directory
@params dir - CString that is the directory to be set to the image file directory
*/
void Controller::C_SetDefaultFileDirectory(const CString& dir)
{
  Settings& settings = m_data.GetSettings();
  settings.SetDefaultFileDir(dir);
  Notify(DEFAULT_DIR_CHANGED);
}

/**
@brief Wrapper function to convert root coordinates
*/
void Controller::C_ConvertOldRoots(const CString& dir, const ImageFileFormat& image_file_format)
{
  Experiment& exp = m_data.GetExperiment();
  exp.ConvertOldRoots(dir, image_file_format);
}

// has progress bar
void Controller::C_ConvertOldRoots(const CString& dir, const ImageFileFormat& image_file_format, CWnd *pParentWnd)
{
  Experiment& exp = m_data.GetExperiment();
  exp.ConvertOldRoots(dir, image_file_format, pParentWnd);
}


// Undo/Redo functions
void Controller::Undo()
{
  if(m_undolist.size() == 0) return;
  else {
    Command *action = m_undolist.back();
    if(action == NULL) return;
    m_redolist.push_back(action);
    m_undolist.pop_back();
    action->Unexecute();
  }
  
  // Currently, the file is changed to unsaved no matter what
  // when you call undo or redo.  Technically, it should only 
  // go to unsaved if the file is different since the last time 
  // saved
  Notify(SAVE_STATUS_UNSAVED);
  
  // test code: comment out or integrate if within these markers --->
  /* 
  int current_command, prev_command;
  bool  duplicate_flag = true;
  if(m_undolist.size() == 0) return;
  
    while (duplicate_flag && m_undolist.size()>0)
    {
    Command *action = m_undolist.back();
    
      m_current_command = action->m_comm_type;
      if (m_current_command == m_prev_command)
      {
      
        
          // if current command type == prev command type 
          // && if they operated on the same thing
          // 
          // push back command to the redo list
          // 
  */
  // <---- end of test code
}


void Controller::Redo()
{
  if(m_redolist.size() == 0) return;
  else {
    Command *action = m_redolist.back();
    if(action == NULL) return;
    m_undolist.push_back(action);
    m_redolist.pop_back();
    action->Execute();
  }
  
  // Currently, the file is changed to unsaved no matter what
  // when you call undo or redo.  Technically, it should only 
  // go to unsaved if the file is different since the last time 
  // saved
  Notify(SAVE_STATUS_UNSAVED);
}

bool Controller::DoCommand(Command *com)
{
  if(com == NULL) return false;
  ClearRedoList();
  if(com->Execute())
  {
    m_undolist.push_back(com);
    return true;
  }
  else return false;
}

void Controller::ClearCommandList(CommandList* pList)
{
  CommandList::iterator it;
  for (it = pList->begin(); it != pList->end(); it++)
    delete *it;
  pList->clear();
}

void Controller::ClearRedoList()
{
  ClearCommandList(&m_redolist);
}

void Controller::ClearUndoList()
{
  ClearCommandList(&m_undolist);
}

void Controller::ClearUndoRedo()
{
  ClearRedoList();
  ClearUndoList();
}

// returns true if undo vector is empty
bool Controller::isUndoVectorEmpty()
{
  return (m_undolist.size() == 0);
}

// returns true if redo vector is empty
bool Controller::isRedoVectorEmpty()
{
  return (m_redolist.size() ==0);
}

//AutoDetect Settings
void Controller::C_SetAutoDetectOptions(bool on_load, bool check_diameter, float diameter_value, bool check_length, 
                                        float length_value, bool check_vertical, float degree_value, bool check_post_process)
{
  Settings& settings = m_data.GetSettings();
  
  settings.SetAutoDetectOptions(on_load, check_diameter, diameter_value, check_length, length_value, check_vertical, degree_value, check_post_process);
}