/* 
* Rootfly:  A program for analyzing and measuring roots. 
* Copyright (c) 2007 Clemson University.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or (at
* your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/





#include "DataStructures.h"
#include "version1_files/RootData.h"
#include "version1_files/RootflyData.h"
#include "version1_files/RootOverTime.h"
#include "version1_files/Session_Array.h"
#include "TestDataStructure.h"
#include "ImageFileFormat.h"
#include "ImageClientConverter.h"

#include <math.h>
#include <algorithm>
#include <sstream>
#include <string>

extern DebugLogFile g_debug_logfile;
extern bool g_collapse_sessions;

//////////////////////////
//Class RPoint functions// 
//////////////////////////


RPoint::RPoint()
  : x(-1), y(-1), window(-1)
{
  
}

RPoint::RPoint(int xx, int yy, int wwindow )
{
  x = xx;
  y = yy;
  window = wwindow;
}

bool RPoint::operator==(const RPoint &other) const
{ 
  return x==other.x && y==other.y && window==other.window;
}

bool RPoint::operator!=(const RPoint &other) const
{
  return ( ! operator==(other) );
}
/////////////////////////////
//Class RDiameter functions// 
/////////////////////////////


// !!! 'r_point' should be renamed 'center'
RDiameter::RDiameter()
  : radius(-1), r_point()
{
  
}

RDiameter::RDiameter( double rad, const RPoint& cen )
{
  radius = rad;
  r_point = cen;
}


bool RDiameter::operator==(const RDiameter &other) const
{
  return (radius == other.radius) && (r_point == other.r_point); 
}

bool RDiameter::operator!=(const RDiameter &other) const
{
  return ( ! operator==(other) );
}
////////////////////////////////
//Class RootInstance functions//
////////////////////////////////

/**
* Default constructor 
*/
RootInstance::RootInstance()
  : m_points(),
    m_diameters(),
    m_color(COLOR_WHITE),
    m_condition_code(),
    m_session_number(),
    m_is_alive(true)
{
}

/**
* Default destructor
*/
RootInstance::~RootInstance()
{
}

RPoint* RootInstance::CreatePoint( )
{  
  RPoint pnt;
  m_points.push_back(pnt);
  return &m_points.back();
}

RPoint* RootInstance::InsertPoint(int index)
{
  assert(index >= 0 && index <= m_points.size() - 1);
  
  RPoint point;
  std::vector<RPoint>::iterator it = m_points.begin() + index + 1;
  m_points.insert(it, point);
  return &m_points[index + 1];
}

RPoint* RootInstance::CreateMidPoint( int index )
{
  RPoint pnt;
  std::vector<RPoint>::iterator it = m_points.begin() + index + 1;
  m_points.insert(it, pnt);
  return &m_points[index+1];
}

RDiameter* RootInstance::CreateDiameter()
{
  RDiameter dia;
  m_diameters.push_back(dia);
  return &m_diameters.back();
}


/**
*
* Clears the vectors of points and diameters
* keeps session#,color,condition, and live
* @returns: true on success, false on fail.
*/
bool RootInstance::ClearAll()
{
  m_points.clear();
  m_diameters.clear();
  
  Color m_color = COLOR_WHITE;
  m_condition_code = "";
  m_session_number = -1;
  m_is_alive = false;
  
  return m_points.size() == 0 && m_diameters.size() == 0;
}


/**
@return: true if successful, false if failure
*/
bool RootInstance::DeletePoint(int index)
{
  if(index < 0 || index > m_points.size() - 1 || m_points.size() < 3)
  {
    return false;
  }
  
  m_points.erase(m_points.begin() + index);
  
  //took out error handling...erase NEVER fails
  
  return true;
}

/**
@return: true if successful, false if failure
*/

bool RootInstance::MovePoint(int index, int newX, int newY)
{
  if(index < 0 || index > m_points.size() - 1)
    return false;   // !!! is this necessary?
  assert(index >= 0 && index <= m_points.size() - 1);
  m_points[index].x = newX;
  m_points[index].y = newY;
  return true;
}

/**
@return: true if successful, false if failure
*/
bool RootInstance::ChangeWindowOfPoint(int index, int newWindow)
{
  assert(index >= 0 && index < m_points.size());
  m_points[index].window = newWindow;
  return true;
} 

/**
@return: true if successful, false if failure
*/
bool RootInstance::DeleteDiameter(int index)
{
  assert(index >= 0 && index <= m_diameters.size() - 1);
  
  m_diameters.erase(m_diameters.begin() + index);
  
  //took out error handling...erase does not fail
  return true;
}

/**
@return: true if successful, false if failure
*/
bool RootInstance::MoveDiameter(int index, int x, int y)
{
  
  if (index < 0 || index > m_diameters.size() - 1)
  {
    return false;
  }
  
  m_diameters[ index ].r_point.x = x;
  m_diameters[ index ].r_point.y = y;
  
  return true;
}  

/**
@return: true if successful, false if failure
*/
bool RootInstance::ChangeWindowOfDiameter(int index, int w)
{
  if(index < 0 || index > m_diameters.size() - 1)
  {
    return false;
  }
  
  m_diameters [ index ].r_point.window = w;
  
  return true;
} 

/**
@return: true if successful, false if failure
*/
bool RootInstance::ResizeDiameter(int index, double radius)
{
  
  if (index < 0 || index > m_diameters.size() - 1)
  {
    return false;
  }
  
  m_diameters[ index ].radius = radius;
  
  return true;
}  

/**
* This function will fill a given vector with numbers representing the windows this root instance exists within.
* @param window_list pointer to a vector of ints to be populated with window numbers
* @return Nothing
*/
void RootInstance::GetWindowNumbers(vector<int>* window_list) const
{
  int num_points = m_points.size(), i=0, j=0;
  bool window_recorded = false;
  
  window_list->clear();
  
  //Loop through all points in this root instance.
  for(i=0; i<num_points; i++) 
  {
    //Checks to see if the window for a point is already in the list.
    for(j=0; j<window_list->size(); j++) 
    {
      if(m_points[i].window == window_list->at(j)) 
      {
        window_recorded=true;
      }
    }
    
    //If the window of the point was not in the list, add it.
    if(!window_recorded)
    {
      window_list->push_back(m_points[i].window);
    }
    
    //Reset the window setting for the next loop iteration.
    window_recorded = false;
  }
  
  return;
}

void RootInstance::SetColor(Color color)
{
  switch(color)
  {
  case COLOR_WHITE:
  case COLOR_LIGHTBROWN:
  case COLOR_DARKBROWN:
  case COLOR_BLACK:
    m_color = color;
    break;
  default:
    m_color = COLOR_WHITE;
  }
}

/**
* Returns the length of the current root in a given window
* @param window Window to find length information within
* @param conversion The pixel-to-millimeter conversion ratio from the application.
* @return double which represents length between points in the given window.  -1 indicates that failure occurred, possibly because no points were located in the given window.
*/
const double RootInstance::FindLengthInWindow(int window, float conversion) const
{
  int i,num_points;
  double length = 0.;
  vector<RPoint> points_in_window;
  
  //Error out if the window number is invalid
  if(window < 0) 
  {
    return -1;
  }
  
  if(conversion <= 0)
  {
    return -1;
  }
  
  num_points = m_points.size();
  
  //If a point is in the window, add it to a vector for later processing
  for(i=0; i<num_points; i++) 
  {
    if(m_points[i].window == window) 
    {
      points_in_window.push_back(m_points[i]);
    }
  }
  
  //Error out if no points were found.
  if(points_in_window.empty()) 
  {
    return -1;
  }
  
  //Compute the distance between all points in the given window.
  for(i=0; i<points_in_window.size()-1; i++) 
  {
    length += sqrt(pow((float)points_in_window[i].x - points_in_window[i+1].x, 2) + pow((float)points_in_window[i].y - points_in_window[i+1].y,2));
  }
  
  return length/conversion;
}

/**
* Returns the average diameter, currently there is only one diameter.  This function is ready for multiple diameters.
* @param conversion The pixel-to-millimeter conversion ratio from the application.
* @return double representing (total diameter size)/(number of diameters) / conversion, and -1 if there is an error.
*/
const double RootInstance::GetAverageDiameter(float conversion) const
{
  int i=0;
  double diameter=0;
  
  assert(m_diameters.size() > 0);
  
  assert(m_diameters.size() == 1);  // for now Rootfly only allows one diameter per root, but if this changes in the future then remove this assert
  
  //  //This should never happen.  If it does, there is a bug.
  //  if (m_diameters.size() == 0)
  //  {
  //    return -1;
  //  }
  
  if(conversion <= 0)
  {
    return -1;
  }
  
  //Get a sum of the diameters, used to create a simple average.
  for (i=0; i<m_diameters.size(); i++) 
  {
    diameter += (m_diameters[i].radius * 2);
  }
  
  return (diameter/m_diameters.size()) / conversion;
}

bool RootInstance::operator==(const RootInstance& other) const
{
  if(m_points != other.m_points) return false;
  if(m_diameters != other.m_diameters) return false;
  if(m_color != other.m_color) {return false;}
  if(m_condition_code != other.m_condition_code) {return false;}
  if(m_session_number != other.m_session_number) {return false;}
  if(m_is_alive != other.m_is_alive) {return false;}
  return true;
}

bool RootInstance::operator!=(const RootInstance& other) const
{
  return( !operator==(other));
}

RootInstance& RootInstance::operator=(const RootInstance& original)
{
  SetColor(original.GetColor());
  SetConditionCode(original.GetConditionCode());
  SetSessionNumber(original.GetSessionNumber());
  if(original.IsAlive()) SetAlive();
  else SetDead();
  m_points = original.m_points;
  m_diameters = original.m_diameters;
//
//  const std::vector <RPoint> *points;
//  const std::vector <RDiameter> *diameters;
//  points = original.GetAllPoints();
//  diameters = original.GetAllDiameters();
//  for( int i = 0; i < points->size(); i++ ){
//    RPoint *point = CreatePoint();
//    *point = (*points)[i];
//  }
//  for( int j = 0; j < diameters->size(); j++ ){
//    RDiameter *diam = CreateDiameter();
//    *diam = (*diameters)[j];
//  }
//  
  return *this;
}
////////////////////////
//Class Root Functions//
////////////////////////
CString Root::DEFAULT_CONDITION_CODE = "NO_CODE";
RootInstance::Color Root::DEFAULT_ROOT_COLOR = RootInstance::COLOR_DARKBROWN;
/**
*
* Constructor
*/
Root::Root( int id ) : m_id(id), m_window_id(0), m_primary_window(-1)
{}

Root::Root() : m_id(-1), m_window_id(-1), m_primary_window(-1)
{}

/**
* Deconstructor
*/
Root::~Root()
{
}

//Assignment Operator
Root& Root::operator=(const Root& original)
{
  SetID( original.GetID() );
  SetPrimaryWindow( original.GetPrimaryWindow() );
  SetDisplayableID( original.GetDisplayableID() );

  m_instances.clear();
  std::vector<RootInstance>::const_iterator iter;
  for(iter = original.GetAllRootInstances()->begin() ; iter < original.GetAllRootInstances()->end(); iter ++)
  {
    RootInstance *inst = CreateRootInstance();
    *inst = *iter;
  }
  return *this;
}

/**
* Creates a new session for the root and returns a reference to it.
*/
RootInstance* Root::CreateRootInstance(int session_number)
{
  
  // checks to see if root instance with same session number already exists
  // if it does, it returns existing instance
  for (size_t ix=0 ; ix < m_instances.size(); ++ix)
  {
    if (m_instances[ix].GetSessionNumber() == session_number)
    {
      return &m_instances.at(ix);
    }
  }
  
  RootInstance new_root_instance;
  new_root_instance.SetSessionNumber(session_number);
  new_root_instance.SetColor(DEFAULT_ROOT_COLOR);
  new_root_instance.SetConditionCode(DEFAULT_CONDITION_CODE);
  
  m_instances.push_back(new_root_instance);
  std::sort(m_instances.begin(),m_instances.end());
  return GetRootInstance(session_number);
}

RootInstance* Root::CreateRootInstance()
{
  RootInstance new_root_instance;  
  RootInstance* return_value;
  new_root_instance.SetColor(DEFAULT_ROOT_COLOR);
  new_root_instance.SetConditionCode(DEFAULT_CONDITION_CODE);
  
  m_instances.push_back(new_root_instance);
  return_value = &m_instances.back();
  std::sort(m_instances.begin(),m_instances.end());
  return return_value;
}

/**
* Removes a root session at position session_number in the m_session vector.
* @param session_number
*/
bool Root::DeleteRootInstance(int session_number)
{
  //This makes much more sense!
  std::vector<RootInstance>::iterator iter = m_instances.begin();
  while(iter != m_instances.end())
  {
    if(iter->GetSessionNumber() == session_number)
    {
      m_instances.erase(iter);
      return true;
    }
    iter++;
  }
  return false;
}


/**
* Copies a root instance to a new session number.
* @param from int representing the source session number to copy from
* @param to int representing the destination session number that will be set in the root instance object.
* @param notified pointer to boolean which keeps track of whether the user has already been notified they will be overwriting roots
* @param previous_choice pointer to an integer representing the users past choice given the previous overwrite notification (IDOK or IDCANCEL)
* @return bool indicating success(true) or failure(false)
*/
bool Root::CopyRootInstance(int from, int to)
{
  //int i=0, j=0, from_index=-1, to_index=-1;
  
  // check input parameters
  assert(from >= 1 && to >= 1);
  if (from < 1 || to < 1 || from == to)  return false;  // invalid session numbers
  if (m_instances.size() == 0)  return false;  // there's nothing in the list, so we can't copy previous
  
  //This iterator is used in case we need to delete something from the m_instances vector
  vector<RootInstance>::iterator iter = m_instances.begin();
  vector<RootInstance>::iterator to_root2;
  RootInstance *from_root, *to_root;
  from_root = NULL;
  to_root = NULL;
  
  //Search for the source root instance, and set the new one equal to it.  Also find the destination index for replacement purposes.
  while(iter != m_instances.end())
  {
    if (iter->GetSessionNumber() == from)  
    { 
      assert(from_root == NULL);  // should not be more than one instance in a session
      from_root = static_cast<RootInstance*>( &iter[0] );
    }
    if (iter->GetSessionNumber() == to)
    {
      assert(to_root == NULL);  // should not be more than one instance in a session
      to_root = static_cast<RootInstance*>( &iter[0] );
      to_root2 = iter;
    }
    iter++;
  }
  
  // No root instance in the source session was found, so return
  if (from_root == NULL)  return false;
  
  //Create a copy of the "previous" root instance
  RootInstance new_root_instance = *from_root;
  new_root_instance.SetSessionNumber(to);
  
  //If roots already exist in the destination session, ask if the user wants to overwrite them
  if (to_root != NULL )
  {
    
    //Erase the root if the user selected OK.
    m_instances.erase(to_root2);
  } 
  else 
  {
    //iter = m_instances.end();
  }
  
  //This inserts the root into its correct location in the m_instances vector
  //m_instances.insert(iter, 1, new_root_instance);
  m_instances.push_back(new_root_instance);
  
  return true;
}


/**
* Returns a pointer to a RootInstance with the given session_number
* @param session_number
* @return pointer to the root instance for the given session_number
*/
RootInstance* Root::GetRootInstance(int session_number)
{
  for (size_t ix=0 ; ix < m_instances.size(); ++ix)
  {
    if (m_instances[ix].GetSessionNumber() == session_number)
    {
      return &m_instances.at(ix);
    }
  }
  //AfxMessageBox("not found");
  return NULL;
}

/**
* Returns a const pointer to a RootInstance with the given session_number
* @param session_number
* @return const pointer to the root instance for the given session_number
*/
const RootInstance* Root::GetRootInstance(int session_number) const
{
  for (size_t ix=0 ; ix < m_instances.size(); ++ix)
  {
    if (m_instances[ix].GetSessionNumber() == session_number)
    {
      return &m_instances.at(ix);
    }
  }
  //AfxMessageBox(" Root not found");
  return NULL;
}

/**
* Loops through the root instances and returns the highest session number found.
* @return const integer which is the largest session number in each of the root instance objects, -1 if no instances exist.
*/
int Root::GetLastSessionNumber() const
{
  int i, session=0;
  
  //Error out if there are no sessions
  if(m_instances.size() == 0) 
  {
    assert(0);  // This should never happen.  Root is deleted if last instance is deleted.  See bug 1911273.  -- akruzner  
    return -1;
  }
  
  //Loop through and return the *highest* session number.  This will work even if the sessions are out of order.
  for(i=0; i<m_instances.size(); i++) 
  {
    if(m_instances[i].GetSessionNumber() > session)
      session = m_instances[i].GetSessionNumber();
  }
  
  return session;
}

/**
* Returns the first session for this particular root
* @return const int representing the session number, and -1 if there are no instances of this root.
*/
int Root::GetFirstSessionNumber() const
{
  int min;
  if(m_instances.size() == 0) 
  {
    assert(0);  // This should never happen.  Root is deleted if last instance is deleted.  See bug 1911273.  -- akruzner  
    return -1;
  }
  
  min = m_instances[0].GetSessionNumber();
  //Loop through and return the *lowest* session number.  This will work even if the sessions are out of order.
  for(int i=1; i<m_instances.size() ; i++) 
  {
    if(min > m_instances[i].GetSessionNumber()) 
    {
      min = m_instances[i].GetSessionNumber();
    }
  }
  
  return min;
}

/**
* Returns the first session this root showed up dead
* @return const int representing the session number this root was first declared dead, -2 if it's not dead and -1 if there are no instances.
*/
int Root::GetFirstDeadSessionNumber() const
{
  int i;
  int session = ROOT_NEVER_DIED;
  bool found = false;
  
  //Error out if there are no root instances.
  if(m_instances.size() == 0) 
  {
    assert(0);  // This should never happen.  Root is deleted if last instance is deleted.  See bug 1911273.  -- akruzner  
    return -1;
  }
  //Loop through and return the *lowest* dead session number.  This will work even if the sessions are out of order.
  for(i=1; i<m_instances.size(); i++) 
  {
    if(!m_instances[i].IsAlive() && (session > m_instances[i].GetSessionNumber() || session < 0)) 
    {
      session = m_instances[i].GetSessionNumber();
    }
  }
  
  //Will be ROOT_NEVER_DIED unless changed in loop
  return session;
}

/**
* This function sums the length of the root over multiple windows, for each session and returns the largest value.
* @param conversion the conversion ratio to apply to the length.
* @return const double representing the actual maximum length of the root in mm.  -1 if there are no sessions.
*/
double Root::GetMaxLength(float conversion) const 
{
  vector<int> windows;    //Holds the window numbers where this root exists
  int i=0,j=0;
  double length=0.0, templength;
  
  //Error out if there are no instances
  if(m_instances.size() == 0) 
  {
    assert(0);  // This should never happen.  Root is deleted if last instance is deleted.  See bug 1911273.  -- akruzner  
    return -1;
  }
  
  //Loop through all instances, for each one sum the lengths of the root in each window it exists within.  Then
  //  compare each of these "sums" to every other sum finding the largest one for the entire root.
  for(i=0; i<m_instances.size(); i++) 
  {
    templength = 0.0;
    m_instances[i].GetWindowNumbers(&windows);
    
    //Find the sum within each respective window and sum it up.
    for(j=0; j<windows.size(); j++) 
      templength+= m_instances[i].FindLengthInWindow(windows.at(j), conversion);
    
    if(templength > length) 
      length = templength;
  }
  
  return length;
  
}

/**
* Searches over all diameters and returns the largest diameter found.
* @param conversion the mm conversion ratio used to convert the pixel lengths to mm.
* @return const double representing the maximum root diameter found, and -1 if there are no root instances.
*/
double Root::GetMaxDiameter(float conversion) const 
{
  int i=0;
  double diameter=0.0;
  
  //Error out if there are no root instances
  if(m_instances.size() == 0) 
  {
    assert(0);  // This should never happen.  Root is deleted if last instance is deleted.  See bug 1911273.  -- akruzner  
    return -1;
  }
  
  //Loop through every instance, taking each average diameter and keep the largest one found
  //  NOTE:  Currently GetAverageDiameter() simply returns the size of one diameter.  The function
  //          is capable of returning an actual average though.
  for(i=0; i<m_instances.size(); i++) 
  {
    double d = m_instances[i].GetAverageDiameter(conversion);
    if (d > diameter) 
    {
      diameter = d;
    }
  }
  
  return diameter;
}


/**
*
* Clears the vector of instances, m_instances
* 
* @returns: true on success, false on fail.
*/
bool Root::ClearAll()
{
  m_instances.clear();
  
  m_id = -1;
  m_primary_window = -1;
  return m_instances.size() == 0;
}

/**
*
* Equality operator
*
*/
bool Root::operator==( const Root& other ) const
{
  if ( m_id != other.m_id )
  {
    return false;
  }
  if ( m_window_id != other.m_window_id )
  {
    return false;
  }
  if ( m_primary_window != other.m_primary_window )
  {
    return false;
  }
  if ( m_instances != other.m_instances )
  {
    return false;
  }
  
  return true;
}

/**
*
* Inequality operator
*
*/
bool Root::operator!=( const Root& other ) const
{
  return ( ! operator==(other) );
}


////////////////////////
//Class Tube Functions//
////////////////////////


/**
*
* Default constructor
*/
Tube::Tube()
: m_next_root_id(1)
{     
  //AfxMessageBox("Default Tube Constructor");
}

/**
*
* Constructor that sets the tag with Cstring tag argument. Each physical tube can have multiple transects.
* For simplicity the transects are just referenced as another tube.
* @param tag - the tag to name the tube
*/
Tube::Tube(const CString& tag)
: m_next_root_id(1), m_tag(tag)
{
}


/**
*
* Default destructor
*/
Tube::~Tube()
{
  
}

/*
*
* Comparison Operator (Logical Equals)
* 
*/
bool Tube::operator==(const Tube &other) const
{
  int i;
  bool RetValue = true;
  
  //Check to see if ImageData vector sizes are the same
  if(other.m_image_scanned.size() != m_image_scanned.size())
  {
    RetValue = false;
  }
  
  for(i = 0; i < m_image_scanned.size(); i++)
  {
    if( m_image_scanned[i] != other.m_image_scanned[i] )
    {
      RetValue = false;
    }
  }
  
  //Check to see if Root vectors have the same size
  if(m_root.size() != other.m_root.size())
  {
    RetValue = false;
  }
  if(m_roots_in_window.size() != other.m_roots_in_window.size())
  {
    RetValue = false;
  }
  
  for(i = 0; i < m_root.size(); i++)
  {
    if(!(m_root[i] == other.m_root[i]))
    {
      RetValue = false;
    }
  }
  
  for(i = 0; i < m_roots_in_window.size(); i++)
  {
    if(!(m_roots_in_window[i] == other.m_roots_in_window[i]))
    {
      RetValue = false;
    }
  }
  
  if(m_tag != other.m_tag)
  {
    RetValue = false;
  }
  if(m_next_root_id != other.m_next_root_id) 
  {
    RetValue = false;
  }
  
  if(m_image_annotated.size() != other.m_image_annotated.size()) 
  {
    RetValue = false;
  }
  for(i = 0; i < m_image_annotated.size(); i++) 
  {
    if(m_image_annotated[i].session != other.m_image_annotated[i].session)
    {
      RetValue = false;
    }
    if(m_image_annotated[i].window != other.m_image_annotated[i].window) 
    {
      RetValue = false;
    }
  }
  
  if(m_image_exists.size() != other.m_image_exists.size()) 
  {
    RetValue = false;
  }
  for(i = 0; i < m_image_exists.size(); i++) 
  {
    if(m_image_exists[i].session != other.m_image_exists[i].session) 
    {
      RetValue = false;
    }
    if(m_image_exists[i].window != other.m_image_exists[i].window)
    {
      RetValue = false;
    }
  }
  
  return RetValue;
  
}

bool Tube::operator!=(const Tube &other) const
{
  return ( ! (*this == other) );
}

/**
*
* Clears the vector of roots
* keeps tag the same.
* @returns: true on success, false on fail.
*/
bool Tube::ClearAll()
{
  m_root.clear();
  
  m_tag = "";
  return m_root.size() == 0;
}

/**
* Creates a new root and adds it to the m_root vector
* @returns: address of newly created root
*/
Root* Tube::CreateRoot()
{  
  
  Root new_root( m_next_root_id );
  m_next_root_id++;
  
  m_root.push_back(new_root);
  return &m_root.back();
}


/**
*
* Creates a new root and adds it to the m_root vector
* @params: window - current window number to add root to
* @returns: address of newly created root
*/
Root* Tube::CreateRoot(int w)
{
  Root new_root( m_next_root_id++); 
  new_root.SetPrimaryWindow(w);
  new_root.SetDisplayableID(GetNextRootWindowID(w));
  m_root.push_back(new_root);
  return &m_root.back();  
}

/**
*
* Creates a new root given a root id and adds it to the m_root vector
* used for undo/redo functions that already have root_id to associate
* @params: window - current window number to add root to
* @returns: address of newly created root
*/
Root* Tube::CreateRoot(int w, int root_id)
{
  Root new_root( root_id ); 
  new_root.SetPrimaryWindow(w);
  new_root.SetDisplayableID(GetNextRootWindowID(w));
  m_root.push_back(new_root);
  return &m_root.back();  
}


/**
*
* Removes a root with specified position root_number from m_root
* @param root_number - the root position
* @returns: bool in\dicating success (0 good, 1 failure)
*/
bool Tube::DeleteRoot(int root_id)
{
  vector<Root>::iterator r_ptr;
  
  for( r_ptr = m_root.begin(); r_ptr != m_root.end(); r_ptr++ )
  {
    if( r_ptr->GetID() == root_id)
    {
      m_root.erase(r_ptr);
      break;
    }
  }
    
  return true;
}

/**
*
* set tube with specified tag
* @param tag - the name of the tube
*/
void Tube::SetTubeTag(const CString& tag)
{
  m_tag = tag;
}

/**
*
* get tube tag
* @return CString
*/
CString Tube::GetTag() const
{
  return m_tag;
}

/**
* gets the root of specified ID
* @return address of root
*/
Root* Tube::GetRoot(int id)
{
  if ( id < 1 || id >= m_next_root_id) return NULL;
  
  int i = 0;
  vector<Root>::iterator iter;
  for(iter = m_root.begin(); iter != m_root.end(); iter++)
  {
    if(iter->GetID() == id) return &m_root[i];
    i++;
  }
  
  return NULL;
}

const Root* Tube::GetRoot(int id) const
{
  if (id < 1 || id >= m_next_root_id) return NULL;
  
  vector<Root>::const_iterator iter;
  
  for(iter = m_root.begin(); iter != m_root.end(); iter++)
  {
    if(iter->GetID() == id) return static_cast<const Root*>( &(*iter) );
  }
  return NULL;
}

/**
* Given a window number, this will fill a vector with root objects and return it to the user
* @param window The window number to search for roots within
* @return pointer to a vector of roots, NULL if window is invalid
*/

std::vector<Root>* Tube::GetRootsInWindow(int window)
{
  assert(window >= 1);
  
  //If the function received an invalid window number, return NULL
  if(window < 1)
    return NULL;
  
  //If the vector is not empty, clear it.
  if(m_roots_in_window.size() > 0)
    m_roots_in_window.clear();
  
  
  for(int i = 0; i < m_root.size(); i++)
  {
    if(m_root[i].GetPrimaryWindow() == window)
      m_roots_in_window.push_back(m_root[i]);
  }
  
  return &m_roots_in_window;
}

int Tube::GetNumberOfRootsInWindowSession(int window, int session) const
{
  std::vector<Root> roots;
  const RootInstance *inst;
  int count = 0;
  
  if(window < 1) return 0;
  for(int i = 0; i < m_root.size(); i++)
  {
    if( (m_root[i].GetPrimaryWindow() == window) && (inst = m_root[i].GetRootInstance(session)) ) count++;
  }
  return count;
}

int Tube::GetNextRootWindowID(int primary_window)
{
  assert(primary_window >= 1);
  int max_id = 0;
  if(primary_window < 1)    return -1;
  for( int i = 0; i < m_root.size(); i++ )
  {
    if( m_root[i].GetPrimaryWindow() == primary_window )
    {
      int id = m_root[i].GetDisplayableID();
      if (id > max_id)  max_id = id;
    }
  }
  return max_id + 1;
}

// returns whether or not imaged for the
// current tube window and session has been scanned by the automatic detector
const bool Tube::GetImageScanned(int window, int session) const
{
  for(int i = 0; i < m_image_scanned.size(); i++)
  {
    if(m_image_scanned[i].window == window && m_image_scanned[i].session == session)
    {
      return true;
    }
  }
  return false;
}

// sets image refrenced by window and session is scanned
// or removes the label depending on status
void Tube::SetImageScanned(bool status, int window, int session)
{
  vector<ImageData>::iterator v1;
  if(status)
  {
    // size isn't zero, check to see if already labeled 
    if(m_image_scanned.size() != 0)
    {
      v1 = m_image_scanned.begin();
      while(v1 != m_image_scanned.end())
      {
        // already labeled as scanned, do nothing
        if((v1->window == window) && v1->session == session) return;
        v1++;
      }      
    }
    // not in vector, place at end
    ImageData temp(window,session);     
    m_image_scanned.push_back(temp);
  } 
  else // remove an instance of a scanned image
  {
    // size isn't zero, check to see if already labeled 
    if(m_image_scanned.size() != 0)
    {
      for(v1 = m_image_scanned.begin();
      v1 != m_image_scanned.end(); ++v1)
      {
        if((v1->window == window) && (v1->session == session))
        {
          m_image_scanned.erase(v1);
          return;
        }
      }
    }
  }
}



const bool Tube::GetImageAnnotated(int window, int session) const
{
  if(window < 0 || session < 0)
    return false;
  for(int i = 0; i < m_image_annotated.size(); i++)
  {
    if(m_image_annotated[i].window == window && m_image_annotated[i].session == session)
    {
      return true;
    }
  }
  return false;
}

int Tube::GetImageNumAnnotated() const
{
  return m_image_annotated.size();
}

// sets image refrenced by window and session is scanned
// or removes the label depending on status
void Tube::SetImageAnnotated(bool status, int window, int session)
{
  
  if(status)
  {
    // size isn't zero, check to see if already labeled 
    if(m_image_annotated.size() != 0)
    {
      
      for(vector<ImageData>::iterator it = m_image_annotated.begin();
      it != m_image_annotated.end(); ++it)
      {
        if((it->window == window) && (it->session == session))
        {
          return;
        }
      }
    }
    // not in vector, place at end
    ImageData temp(window,session);     
    m_image_annotated.push_back(temp);
  } else // remove an instance of a scanned image
  {
    // size isn't zero, check to see if already labeled 
    if(m_image_annotated.size() != 0)
    {
      for(vector<ImageData>::iterator it = m_image_annotated.begin();
      it != m_image_annotated.end(); ++it)
      {
        if((it->window == window) && (it->session == session))
        {
          m_image_annotated.erase(it);
          return;
        }
      }
    }
  }
}

const bool Tube::GetImageExists(int window, int session) const
{
  for(int i = 0; i < m_image_exists.size(); i++)
  {
    if(m_image_exists[i].window == window && m_image_exists[i].session == session)
    {
      return true;
    }
  }
  return false;
}

int Tube::GetImageNumExists() const
{
  return m_image_exists.size();
}

// sets image refrenced by window and session is scanned
// or removes the label depending on status
void Tube::SetImageExists(bool status, int window, int session)
{
  vector<ImageData>::iterator it;
  if(status)
  {
    // size isn't zero, check to see if already labeled 
    if(m_image_exists.size() != 0)
    {
      
      for(it = m_image_exists.begin();
      it != m_image_exists.end(); ++it)
      {
        if((it->window == window) && (it->session == session))
        {
          return;
        }
      }
    } 
    // not in vector, place at end
    ImageData temp(window,session);     
    m_image_exists.push_back(temp);
  } else // remove an instance of a scanned image
  {
    // size isn't zero, check to see if already labeled 
    if(m_image_exists.size() != 0)
    {
      for(it = m_image_exists.begin();
      it != m_image_exists.end(); ++it)
      {
        if((it->window == window) && (it->session == session))
        {
          m_image_exists.erase(it);
          return;
        }
      }
    }
  }
}

/**
*
* A constructor
*/
Experiment::Experiment()
  : m_tubes(),
    m_session_dates(),
    m_image_file_format(),
    m_image_dir(),
    m_mm_conversion(1.0),  // By default, the experiment mm conversion is 1.0
    m_min_windows(0),
    m_min_sessions(0),
    m_max_windows(0),
    m_max_sessions(0),
    m_is_old(false)   // By default the experiment is not an old file convert
{
}

/**
*
* A deconstructor
*/
Experiment::~Experiment()
{
  
}


/**
*
* @returns m_tubes - number of tubes in experiment
*/
int Experiment::GetNTubes() const
{
  return m_tubes.size();
}


/**
*
* Searches vector for matching tag. Because the tag could potentially return NULL, a pointer is used
* instead of a reference.
* @returns: address of tube
*/
Tube* Experiment::GetTubeByName(const TubeName& tag)
{
  for (size_t ix=0 ; ix < m_tubes.size(); ++ix)    
  {                                             
    if (m_tubes[ix].GetTag() == tag)            
    {
      return &m_tubes.at(ix);
    }
  }
  return NULL;
}


/**
*
* Searches vector for matching tag. Because the tag could potentially return NULL, a pointer is used
* instead of a reference.
* @returns: address of tube
*/
const Tube* Experiment::GetTubeByName(const TubeName& tag) const
{                                              
  for (size_t ix=0 ; ix < m_tubes.size(); ++ix)
  {
    if (m_tubes[ix].GetTag() == tag)
    {
      return &m_tubes.at(ix);
    }
  }
  
  return NULL;
}


/* Checks to see if two experiments are the same
* 
*/
bool Experiment::operator==(const Experiment& other) const
{
  int i;
  
  if ( m_tubes.size() != other.m_tubes.size() )  return false;
  if ( m_session_dates.size() != other.m_session_dates.size() )  return false;
  for(i = 0; i < m_tubes.size(); i++)
  {
    if(m_tubes[i] != other.m_tubes[i]) return false;
  }
  
  for(i = 0; i < m_session_dates.size(); i++)
  {
    if( !(m_session_dates[i] == other.m_session_dates[i]) ) return false;
  }
  
  if(m_image_dir != other.m_image_dir) return false;
  
  if( !(m_image_file_format == other.m_image_file_format) ) return false;
  if(m_mm_conversion != other.m_mm_conversion) return false;
  if(m_min_windows != other.m_min_windows) return false;
  if(m_max_windows != other.m_max_windows) return false;
  if(m_min_sessions != other.m_min_sessions) return false;
  if(m_max_sessions != other.m_max_sessions) return false;
  if(m_is_old != other.m_is_old) return false;
  
  return true;
  
}


/**
*
* Creates a tube with the specified tag
* @param tag - the name of the tube
* @returns: address of newly created tube
*/
Tube* Experiment::CreateTube(const CString& tag)
{
  Tube new_tube(tag);
  m_tubes.push_back(new_tube);
  return &m_tubes.back();
}


/**
*
* Removes tube with specified tag in the m_tubes vector
* @param tag - the name of the tube
* @returns: bool indicating success
*/
bool Experiment::DeleteTube(const CString& tag)
{
  
  for (size_t ix=0 ; ix < m_tubes.size(); ++ix)
  {
    if (m_tubes[ix].GetTag() == tag)
    {
      m_tubes.erase(m_tubes.begin()+ix);
      return true;
    }
  }
  
  return false;
}


bool Experiment::SetImageDir( const CString& directory )
{
  m_image_file_format.SetFileDirectory(directory);
  m_image_dir = directory;   // !!! probably shouldn't store this in experiment, if ImageFileFormat already does -- STB 4/12/10
  return m_image_dir?true:false;  // ?? !!! CString will never be null; no need to return anything -- STB 4/12/10
}

const CString Experiment::GetImageDir() const
{
  return m_image_dir;          
}

bool Experiment::SetMMConversion( float mm_conversion )
{
  m_mm_conversion = mm_conversion;   
  return m_mm_conversion?true:false; 
}

float Experiment::GetMMConversion() const
{
  return m_mm_conversion;      
}


/**
* Loads a Experiment of Tubes into the current data structure.
*
* The file pointer will be located at the beginning of a Tube and the 
* function will load all Roots associated with the Tube.
*
* @author Hamilton Evans
* @param *filename the filename to be opened and loaded.
* @param *parameters a pointer to a Settings object.
* @return Success (0) or Failure (one of the nonzero ERRORS)
*/
int Experiment::Load(const char *filename, Settings *parameters)
{  
  int           num_tubes, i;
  char          version[20], tempBuf[1000];
  float         conversion;
  CString       image_dir = " ", tube_tag = " ";
  FILE          *fpt;
  Tube          *new_tube; 
  
  fpt = fopen(filename, "rt");  // !!! text format is used; will be incompatible on Unix system
  
  /// If the file cannot be opened, the function will return -1.
  if(fpt == NULL) 
  { 
    return ERRORS::RFY_LOADERROR_NOFILEFOUND;
  }
  
  g_debug_logfile.Print("Experiment::Load() 1 %s %x\n", filename, parameters);
  
  // check to make sure that a valid rootfly file is being opened
  // first line must read Rootfly_XX although only up to the _ is checked
  CString file_version = fgets(version, 9, fpt);
  if (file_version != "Rootfly_")
  {
    fclose(fpt);
    return ERRORS::RFY_LOADERROR_FILECORRUPT;
  }
  
  //Clear Experiment
  if(!Experiment::ClearAll())
  {
    fclose(fpt);
    return ERRORS::RFY_LOADERROR_FILECORRUPT;
  }
  
  // Get the version number in the file
  fscanf(fpt,"%20s\n",&version);  
  
  g_debug_logfile.Print("Experiment::Load() 2\n");
  
  // If the file is an old Rootfly file ("Rootfly_11"), use Experiment::LoadOld()
  if (strcmp(version,"11") == 0)
  {
    g_debug_logfile.Print("Experiment::Load() 3a\n");
    
    fclose(fpt);
    DWORD ret = AfxMessageBox("File was created with an old version of Rootfly.  Convert to the new format?", MB_YESNO);
    if (ret != IDYES)  
    {
      return ERRORS::RFY_LOADERROR_USERCANCELED;
    }
    // Set old file to true
    m_is_old = true;
    int status = LoadOld(filename, parameters);
    return status;
  }
  
  // If the file is a new Rootfly file ("Rootfly_20"), use Experiment::LoadOld()
  else if ( (file_version + version) == parameters->GetVersion())
  {
    g_debug_logfile.Print("Experiment::Load() 3b\n");
    
    // Current experiment is NOT an old converted file
    SetOldFile(false);
    
    //Read the image directory with spaces
    char *status;
    status = fgets ( tempBuf, 1000, fpt);
    tempBuf[strlen(tempBuf)-1] = '\0';
    image_dir = tempBuf;
    
    g_debug_logfile.Print("Experiment::Load() 4\n");
    
    if(status == NULL)
    {
      fclose(fpt);
      return ERRORS::RFY_LOADERROR_FILECORRUPT;
    }
    
    g_debug_logfile.Print("Experiment::Load() 5\n");
    
    
    bool ret_val = m_image_file_format.Load(fpt);
    
    g_debug_logfile.Print("Experiment::Load() 6\n");
    
    if(!ret_val)
    {
      fclose(fpt);
      return ERRORS::RFY_LOADERROR_FILECORRUPT;
    }
    
    g_debug_logfile.Print("Experiment::Load() 7\n");
    
    // loads session dates
    int max_sessions;
    int temp_int;
    int j;
    Date temp_session;
    CString message;
    fscanf(fpt,"%d\n",&max_sessions);
    
    g_debug_logfile.Print("Experiment::Load() 8 maxsessions=%d\n", max_sessions);
    
    // This clears session dates before adding anything -
    // This fixes bug 2893528, which caused the user who used File->Open while an experiment
    // was opened to have their sessions go over the maximum because they kept
    // appending to m_session_dates. Clearing it prevents this issue - JCL
    m_session_dates.clear();
    
    for(i = 1; i <= max_sessions; i++)
    {
      bool match = false;
      fscanf(fpt,"%d %d/%d/%d\n", &temp_int, &temp_session.month,
        &temp_session.day, &temp_session.year);

      // This loop discards duplicate dates.  Only needed for files messed up
      // by bug 2893528. - JCL 4/14/10
      for(j = 0 ; j < m_session_dates.size() ; j++)
      {
        if( m_session_dates[j] == temp_session)
        {
          match = true;
          break;
        }
      }
      g_debug_logfile.Print("Experiment::Load() 8b %d %d/%d/%d match=%d\n", temp_int, temp_session.month, temp_session.day, temp_session.year, match);
      if (!match) 
      {
        m_session_dates.push_back(temp_session);
      }
    }
    
    g_debug_logfile.Print("Experiment::Load() 9\n");
    
    if(fscanf(fpt,"%f\n%d\n\n",&conversion,&num_tubes) == EOF)
    {
      fclose(fpt);
      return ERRORS::RFY_LOADERROR_FILECORRUPT;
    } 
    
    // Set internal variables
    
    m_image_dir = image_dir;
    m_mm_conversion = conversion;
    
    
    g_debug_logfile.Print("Experiment::Load() 10 conv=%8.4f, ntubes=%d\n", conversion, num_tubes);
    
    // Load all Tubes, this will recursively load entire data structure.
    for(i = 0; i < num_tubes; i++)
    {
      status = fgets ( tempBuf, 1000, fpt);
      tempBuf[strlen(tempBuf)-1] = '\0';
      tube_tag = tempBuf;
      
      if(status == NULL)
      {
        fclose(fpt);
        return ERRORS::RFY_LOADERROR_FILECORRUPT;
      }
      
      g_debug_logfile.Print("Experiment::Load() 11 tube %d\n", i);
      
      new_tube = (Tube*) CreateTube( tube_tag );        
      
      g_debug_logfile.Print("Experiment::Load() 12\n");
      
      new_tube->Load(fpt);     
      
      g_debug_logfile.Print("Experiment::Load() 13\n");
    }
  }
  
  else
  {
    g_debug_logfile.Print("Experiment::Load() 14\n");
    
    // unrecognized rootfly file version number
    fclose(fpt);
    return ERRORS::RFY_LOADERROR_VERSION;
  }
  
  g_debug_logfile.Print("Experiment::Load() 15\n");
  
  fclose(fpt);
  return ERRORS::RFY_LOADERROR_SUCCESS;
}

/**
* Saves information from the current datastructure.
* So far the settings are not visible to this object, this will need to change.
*
* @param fp pointer to an open file handle.
* @param *parameters pointer to a Settings object
* @return Success (0) or Failure (-1)
*/

int Experiment::Save(const char *filename, const Settings *parameters) const
{
  if(!filename) return -1;
  g_debug_logfile.Print("Experiment::Save() 1 %s %x\n", filename, parameters);
  FILE *fp = fopen(filename, "wt");  // !!! text format is used; will be incompatible on Unix system
  g_debug_logfile.Print("Experiment::Save() 2 %d\n", fp);
  CString version   = parameters->GetVersion();
  unsigned int num_tubes = m_tubes.size();
  unsigned int i;
  
  
  if(fp == NULL) 
  {
    return -1;
  }
  
  g_debug_logfile.Print("Experiment::Save() 3 %s %s\n", version, m_image_dir);
  if (fprintf(fp,"%s\n%s\n",version,m_image_dir) < 0) return -1;
  
  g_debug_logfile.Print("Experiment::Save() 4\n");
  m_image_file_format.Save(fp);
  g_debug_logfile.Print("Experiment::Save() 5\n");
  
  // write date to file
  g_debug_logfile.Print("Experiment::Save() 6 %d <= %d\n", m_max_sessions, m_session_dates.size());
  {
    int foo = m_session_dates.size();
    //assert(m_max_sessions == m_session_dates.size());  // we know this is failing, but why, and is it important? -- stb 10/17/07
    // Looks like m_session_dates is wrong.  It has 41 elements, but there are only 12 sessions.
    // Perhaps this is the result of loading an .rfy from one computer on another computer (dates from images merged?) -- stb 6/3/09
  }
  fprintf(fp,"%d\n", m_session_dates.size());
  g_debug_logfile.Print("Experiment::Save() 7\n");
  for(i = 0; i < m_session_dates.size() ; i++)
  {
    g_debug_logfile.Print("Experiment::Save() 7b %d\n", i);
    fprintf(fp,"%d %d/%d/%d\n",i,m_session_dates[i].month,
      m_session_dates[i].day, m_session_dates[i].year);
  }
  g_debug_logfile.Print("Experiment::Save() 8\n");
  
  // Write image format string
  //stringstream ss;
  //ImageFileFormatInfo iffi = m_image_file_format.GetImageFileFormatInfo();
  
  //char image_format[1024];
  //sprintf(image_format,"%d %s %s",iffi.numElementsInFileName,iffi.filetype,iffi.delim);
  
  //ss << iffi.numElementsInFileName << " " << iffi.filetype << " " << iffi.delim;
  //OrderElement oe;
  
  //for(i = 0; i < 7; i++)
  //{
  //oe = iffi.InfoOrder[i];
  //ss << " " << oe.name << " " << oe.element << " " << oe.prefix << " " << oe.length << " " << oe.included;
  //sprintf(image_format,"%s %s %d %s %d %d",image_format,oe.name,oe.element,oe.prefix,oe.length,oe.included);
  //sprintf(image_format,"%s %d %s %d %d",image_format,oe.element,oe.prefix,oe.length,oe.included);
  //}
  //CString image_format = ss.str().c_str();
  
  //if(fprintf(fp,"%s\n%s\n%s\n%f\n%d\n\n",version, m_image_dir, image_format, m_mm_conversion, num_tubes) < 0)
  //{
  //  return -1;
  //}
  g_debug_logfile.Print("Experiment::Save() 9\n");
  if(fprintf(fp,"%f\n%d\n\n",m_mm_conversion,num_tubes) < 0) return -1;
  assert(num_tubes == m_tubes.size());
  g_debug_logfile.Print("Experiment::Save() 10 %f %d <= %d\n", m_mm_conversion, num_tubes, m_tubes.size());
  
  //Find the number of sessions in the data structures
  for(i=0; i<num_tubes; i++) 
  {
    g_debug_logfile.Print("Experiment::Save() 10b %d\n", i);
    if((m_tubes[i]).Save(fp) == -1) 
    {
      return -1;
    }
    fprintf(fp, "\n\n");
  }
  g_debug_logfile.Print("Experiment::Save() 11\n");
  fclose(fp);
  g_debug_logfile.Print("Experiment::Save() 12\n");
  return 0;
}


/**
*
* Clears the vector of Tubes, m_tubes
* Clears m_image_dir, and m_mm_conversion as well.
* @returns: true on success, false on fail.
*/
bool Experiment::ClearAll()
{
  m_tubes.clear();
  //should also clear the member vars
  
  m_mm_conversion = 1;
  m_image_dir = "";
  
  // Set current experiment as NOT an old converted file
  m_is_old = false;
  
  return m_tubes.size() == 0;
}


/**
*  This function will convert an old data structure into the new
*  structure.  This will return 0 on success
*
* \author Corey Dickerson
* \param old data structure that will be used for conversion
* \param settings of current experiment
* \return returns 0 on success, -1 on failure
*/

int Experiment::ConvertDataStructure(OldData::RootflyData oldStructure, Settings *settings, const CString &filename)
{  
  CString currentVersion = settings->GetVersion();  
  int ntubes,ntransects,nwindows,nsessions;
  int t,w,p,tr,g, rb, s;
  int rootNum = 1;
  int winID = 1;
  CString temp_str;
  OldData::RootList *old_current_rootlist;
  int num_points, num_rootblocks;
  OldData::RootInstance *old_current_rootinstance;
  CPoint offset;
  bool status;
  
  Tube *currentTube;
  Root *currentRoot;
  RootInstance *currentInstance;
  RDiameter *currentDiameter;  
  RPoint currentPoint;  
  RPoint *currentRootPoint;  
  std::vector<CPoint> old_offsets;
  
  // checks if the parameter settings is not null, so that when we set the version number, we know it is valid
  if(settings == NULL)
  {
    return ERRORS::RFY_LOADERROR_FILECORRUPT;
  }
  
  // The next set of if statements set the important settings information, and return -1 if they are set incorrectly.
  
  // sets the image format for the data structure
  // commented b/c no one was using it anyway -- STB 4/15/10
//  if(!SetImageFormat(oldStructure.m_image_format))
//  { 
//    return ERRORS::RFY_LOADERROR_FILECORRUPT;
//  }
  
  // sets the image directory for the current experiment
  if(!SetImageDir(oldStructure.m_image_directory))
  {
    return ERRORS::RFY_LOADERROR_FILECORRUPT;
  }
  
  // sets the mm conversion for the current experiment
  if(!SetMMConversion(oldStructure.m_mm_conversion))
  {
    return ERRORS::RFY_LOADERROR_FILECORRUPT;
  }
  
  // Get number of tubes, transect, windows, and sessions
  oldStructure.m_root.GetDimensions(&ntubes,&ntransects,&nwindows,&nsessions);
  
  // loop through tubes
  for( t = 0; t < ntubes; t++ ) 
  {
    rootNum = 1;
    winID = 1;
    // loop through transects
    for( tr = 0; tr < ntransects; tr++ ) 
    {
      // Create Tube with appropriate tag
      temp_str.Format("%03d",t+1);
      
      currentTube = CreateTube(temp_str);
      
      // loop through windows
      for( w = 0; w < nwindows; w++ ) 
      {   
        // Get a list of all roots in given tube, transect, and window
        old_current_rootlist = oldStructure.m_root.GetRootList(t,tr,w);
        
        // Resize vector to be size of number of sessions
        // so there ARE holes
        old_offsets.resize(nsessions);
        
        // for each offset in all session
        for( g = 0; g < nsessions; g++) 
        {
          // Get offset for this current session
          offset = old_current_rootlist->offset_list.at(g);
          
          //Put offset in new vector at corresponding location
          old_offsets.at(g) = offset;
          
        }
        
        // get number of root blocks("ROOT") in old structure
        num_rootblocks = old_current_rootlist->size(); 
        if( num_rootblocks > 0 ) 
        {          
          // loop through root blocks
          for( rb = 0; rb < num_rootblocks; rb++ ) 
          {
            //Create new Root that will have root instances
            currentRoot = currentTube->CreateRoot(w+1);
            
            // Set root ID for each ROOT
            currentRoot->SetID(rootNum++);
            currentRoot->SetPrimaryWindow(w+1);
            currentRoot->SetDisplayableID(currentTube->GetNextRootWindowID(w+1)-1);     
            
            
            for(s = old_current_rootlist->at(rb).GetFirstSession() - 1; s < old_current_rootlist->at(rb).GetNumInstances()+old_current_rootlist->at(rb).GetFirstSession()-1; s++)
            {
              if(s >= old_offsets.size() )
                break;
              
              old_current_rootinstance = old_current_rootlist->at(rb).GetRootInstance(s,&status);
              num_points = old_current_rootinstance->points.size();
              
              // Create New root Instance  
              currentInstance = currentRoot->CreateRootInstance(s+1);
              currentInstance->SetConditionCode("NO_CODE_GIVEN");
              
              RootInstance::Color color;
              
              switch(old_current_rootinstance->color)
              {
              case 0: color = RootInstance::COLOR_DARKBROWN; break;
              case 1: color = RootInstance::COLOR_LIGHTBROWN; break;
              case 2: color = RootInstance::COLOR_BLACK; break;
              case 3: color = RootInstance::COLOR_WHITE; break;
              }
              
              currentInstance->SetColor(color);
              
              currentInstance->SetColor(color);
              
              //Set status of root instance (Dead or Alive)
              if(status)
                currentInstance->SetAlive();
              else
                currentInstance->SetDead();
              
              // loop through points
              for( p = 0; p < num_points; p++ ) 
              {
                //Create new point
                currentRootPoint = currentInstance->CreatePoint();
                
                //Set values of the points and then add offset                
                currentRootPoint->x = (old_current_rootinstance->points.at(p).x + old_offsets.at(s).x);
                currentRootPoint->y = (old_current_rootinstance->points.at(p).y + old_offsets.at(s).y);
                currentRootPoint->window = w + 1;
              } // end p
              
              // Create a new diameter
              currentDiameter = currentInstance->CreateDiameter();
              
              // Set the radius of the diameter
              currentDiameter->radius = old_current_rootinstance->diameter/ 2;// * diameterFactor;
              
              // Set the location of the diameter
              currentDiameter->r_point.x = (old_current_rootinstance->diameter_center.x );
              currentDiameter->r_point.y = (old_current_rootinstance->diameter_center.y);
              currentDiameter->r_point.window = w + 1;
              
            } // end s
          } // end r
        }
        else 
        {
        } // end else
      } // end w
    } // end tr
  } // end t
  return 0;
}


// Converts coordinates of root points and diameters from an old Rootfly file (version 1.0.3)
// to the new version (2.0).  Note:  The old format stored client coordinates, which was not
// guaranteed to have the same aspect ratio as the image.  The new format stores image coordinates
// instead.
//
// @param dir  Directory containing images
// @param image_file_format  Specifies the format of the image names in the directory
void Experiment::ConvertOldRoots(const CString& dir, const ImageFileFormat& image_file_format)
{
  CString tube_tag, name;
  int window, last_window = -1;
  int session;
  vector<Tube> *tube;
  vector<Tube>::iterator t_iter;
  vector<Root> *root;
  vector<Root>::iterator r_iter;
  vector<RootInstance> *inst;
  vector<RootInstance>::iterator i_iter;
  const vector<RPoint> *point;  
  const RDiameter *diam;
  RPoint conv_point, conv_radius;
  blepo::ImgBgr img;
  
  // Create temporary converter from client to image coordinates corresponding to old version of Rootfly (version 1.0.3).
  // (SetWindow is not called because the window does not matter for our purposes.  
  //  The window is ignored by ClientToImage anyway).
  ImageClientConverter converter;
  { 
    CRect old_screen(0,0,639,482);  // dimensions of image display in old version of Rootfly (version 1.0.3), which could not be resized
    CString name = image_file_format.GetFirstImage(dir);  // grab dimensions of one image, just in case we happen to have roots whose images do not exist (which should not happen)
    int w = 640, h = 480;  // default, just in case we cannot find an image at all
    if (name != "" && blepo::Load(name, &img))  { w = img.Width();  h = img.Height(); }
    
    converter.SetClientRect( old_screen );
    converter.SetImageSize( CSize(w, h) );
  }
  
  tube =  GetAllTubes();
  if (!tube) return;
  
  // for each tube ...
  for( t_iter = tube->begin(); t_iter != tube->end(); t_iter++ )
  {    
    tube_tag = t_iter->GetTag();
    root = t_iter->GetAllRoots();
    if (!root) continue;
    
    // for each root in tube ...
    for( r_iter = root->begin(); r_iter != root->end(); r_iter++ )
    {
      
      //Not Best way but helps with encapulations
      inst = const_cast<std::vector<RootInstance>*>(r_iter->GetAllRootInstances());
      if (!inst) continue;
      
      // for each root instance ...
      for( i_iter = inst->begin(); i_iter != inst->end(); i_iter++ )
      {        
        session = i_iter->GetSessionNumber();
        point = i_iter->GetAllPoints();
        if (!point) continue;
        
        // for each point ...
        for (int pt = 0; pt < point->size(); pt++ )
        {
          // load image for this tube, window, session (if not already loaded)
          // tell converter the dimensions of this image
          window = point->at(pt).window;
          if( window != last_window)
          {
            // get name of file
            name = image_file_format.GetImageFileName(dir,tube_tag,window,session);
            // if a file was found, then load image and set converter based on image size
            if (name != "" && blepo::Load(name,&img))  converter.SetImageSize( CSize(img.Width(),img.Height()) );
          }
          
          // convert from client to image coordinates
          conv_point = converter.ClientToImage( CPoint(point->at(pt).x, point->at(pt).y) );
          i_iter->MovePoint(pt, conv_point.x,conv_point.y); 
          last_window = window;          
        }        
        last_window = -1;          
        
        // convert diameter in the same way
        diam = i_iter->GetDiameter(0);
        window = diam->r_point.window;
        conv_point = converter.ClientToImage( CPoint(diam->r_point.x, diam->r_point.y) );
        conv_radius = converter.ClientToImage( CPoint(diam->r_point.x+diam->radius, diam->r_point.y+diam->radius) );
        i_iter->MoveDiameter(0, conv_point.x, conv_point.y);
        i_iter->ResizeDiameter(0, conv_radius.x - conv_point.x);
        last_window = window;
      }
    }
  }
  
  // convert pixels/mm conversion
  // (Averages x and y.  Ideally these would be the same, but only if the old
  //  format happened to preserve the aspect ratio of the image.)
  conv_point = converter.ClientToImage( CPoint( m_mm_conversion, m_mm_conversion ) );
  m_mm_conversion = (conv_point.x + conv_point.y) / 2.0;
}

void Experiment::ConvertOldRoots(const CString& dir, const ImageFileFormat& image_file_format, CWnd *pParentWnd)
{
  CString tube_tag, name;
  int window, last_window = -1;
  int session;
  vector<Tube> *tube;
  vector<Tube>::iterator t_iter;
  vector<Root> *root;
  vector<Root>::iterator r_iter;
  vector<RootInstance> *inst;
  vector<RootInstance>::iterator i_iter;
  const vector<RPoint> *point;  
  const RDiameter *diam;
  RPoint conv_point, conv_radius;
  blepo::ImgBgr img;
  
  // progress bar creation
  int lower, upper;
  CProgressCtrl progress_bar;
  CRect rect, rect2, rect3;
  CClientDC dc(pParentWnd);
  dc.SetBkColor(::GetSysColor(COLOR_3DFACE));
  pParentWnd->GetWindowRect(rect);
  pParentWnd->ScreenToClient(rect);
  rect2 = CRect(rect.left+(rect.Width()/2)-150, rect.top+(rect.Height()/2)-15, rect.left+(rect.Width()/2)+150, rect.top+(rect.Height()/2)+15);
  rect3 = CRect(rect2.left,rect2.top-20,rect2.right,rect2.bottom-20);
  dc.DrawText("Converting Old Roots To New Format...         ",rect3,0);
  progress_bar.Create(WS_CHILD | WS_VISIBLE | PBS_SMOOTH , rect2, pParentWnd,0x16);
  
  
  // Create temporary converter from client to image coordinates corresponding to old version of Rootfly (version 1.0.3).
  // (SetWindow is not called because the window does not matter for our purposes.  
  //  The window is ignored by ClientToImage anyway).
  ImageClientConverter converter;
  { 
    CRect old_screen(0,0,639,482);  // dimensions of image display in old version of Rootfly (version 1.0.3), which could not be resized
    CString name = image_file_format.GetFirstImage(dir);  // grab dimensions of one image, just in case we happen to have roots whose images do not exist (which should not happen)
    int w = 640, h = 480;  // default, just in case we cannot find an image at all
    if (name != "" && blepo::Load(name, &img))  { w = img.Width();  h = img.Height(); }
    
    converter.SetClientRect( old_screen );
    converter.SetImageSize( CSize(w, h) );
  }
  
  tube = GetAllTubes();
  if (!tube) return;
  
  
  // more progress bar
  lower = 0;
  upper = tube->size();
  progress_bar.SetRange(lower,upper);
  progress_bar.SetStep(1);
  
  // for each tube ...
  for( t_iter = tube->begin(); t_iter != tube->end(); t_iter++ )
  {    
    // progress bar step
    progress_bar.StepIt();
    
    tube_tag = t_iter->GetTag();
    root = t_iter->GetAllRoots();
    if (!root) continue;
    
    // for each root in tube ...
    for( r_iter = root->begin(); r_iter != root->end(); r_iter++ )
    {
      //Not best way but helpes with encapulation
      inst = const_cast<std::vector<RootInstance>*>(r_iter->GetAllRootInstances());
      if (!inst) continue;
      
      // for each root instance ...
      for( i_iter = inst->begin(); i_iter != inst->end(); i_iter++ )
      {        
        session = i_iter->GetSessionNumber();
        point = i_iter->GetAllPoints();
        if (!point) continue;
        
        // for each point ...
        for (int pt = 0; pt < point->size(); pt++ )
        {
          // load image for this tube, window, session (if not already loaded)
          // tell converter the dimensions of this image
          window = point->at(pt).window;
          if( window != last_window)
          {
            // get name of file
            name = image_file_format.GetImageFileName(dir,tube_tag,window,session);            
            // if a file was found, then load image and set converter based on image size
            if (name != "" && blepo::Load(name,&img))  converter.SetImageSize( CSize(img.Width(),img.Height()) );
          }
          
          // convert from client to image coordinates
          conv_point = converter.ClientToImage( CPoint(point->at(pt).x, point->at(pt).y) );
          i_iter->MovePoint(pt, conv_point.x,conv_point.y); 
          last_window = window;          
        }        
        last_window = -1;          
        
        // convert diameter in the same way
        diam = i_iter->GetDiameter(0);
        window = diam->r_point.window;
        conv_point = converter.ClientToImage( CPoint(diam->r_point.x, diam->r_point.y) );
        conv_radius = converter.ClientToImage( CPoint(diam->r_point.x+diam->radius, diam->r_point.y+diam->radius) );
        i_iter->MoveDiameter(0, conv_point.x, conv_point.y);
        i_iter->ResizeDiameter(0, conv_radius.x - conv_point.x);
        last_window = window;
      }
    }
  }
  
  // convert pixels/mm conversion
  // (Averages x and y.  Ideally these would be the same, but only if the old
  //  format happened to preserve the aspect ratio of the image.)
  conv_point = converter.ClientToImage( CPoint( m_mm_conversion, m_mm_conversion ) );
  m_mm_conversion = (conv_point.x + conv_point.y) / 2.0;
}


// Finds the TubeName with the next largest TubeName from tube_name
// Returns the largest TubeName if none are greater than tube_name
// Returns the empty string if no tubes exist
// Returns the largest TubeName if tube_name not found
TubeName Experiment::GetNextTubeName(const TubeName& tube_name) const
{
  TubeName next_tube = GetLastTubeName();
  std::vector<Tube>::const_iterator iter;
  
  // find the Tube with the next largest TubeName 
  for (iter = m_tubes.begin(); iter != m_tubes.end(); iter++)
  {
    if (tube_name < iter->GetTag() && iter->GetTag() < next_tube)
    {
      next_tube = iter->GetTag();
    }
  }
  
  return next_tube;
}

// Finds the TubeName with the next smallest TubeName from tube_name
// Returns the smallest TubeName if none are less than tube_name
// Returns the empty string if no tubes exist
// Returns the smallest TubeName if tube_name not found
TubeName Experiment::GetPrevTubeName(const TubeName& tube_name) const
{
  TubeName prev_tube = GetFirstTubeName();
  std::vector<Tube>::const_iterator iter;
  
  // find the Tube with the next smallest TubeName
  for (iter = m_tubes.begin(); iter != m_tubes.end(); iter++)
  {
    if (prev_tube < iter->GetTag() && iter->GetTag() < tube_name)
    {
      prev_tube = iter->GetTag();
    }
  }
  
  return prev_tube;
}

// Finds the smallest TubeName in the set of Tubes
// Returns the empty string if no tubes exist
TubeName Experiment::GetFirstTubeName() const
{
  if  (m_tubes.size() == 0) return "";
  
  TubeName first_tube = m_tubes.begin()->GetTag();
  std::vector<Tube>::const_iterator iter;
  
  for (iter = m_tubes.begin(); iter != m_tubes.end(); iter++)
  {
    if (iter->GetTag() < first_tube)
    {
      first_tube = iter->GetTag();
    }
  }
  
  return first_tube;
}


// Finds the largest TubeName in the set of Tubes
// Returns the empty string if no tubes exist
TubeName Experiment::GetLastTubeName() const
{
  if  (m_tubes.size() == 0) return "";
  
  TubeName last_tube = m_tubes.begin()->GetTag();
  std::vector<Tube>::const_iterator iter;
  
  for (iter = m_tubes.begin(); iter != m_tubes.end(); iter++)
  {
    if (last_tube < iter->GetTag())
    {
      last_tube = iter->GetTag();
    }
  }
  
  return last_tube;
}


/**
*  This function will load the old file format into the
*  old data structure.  That data structure will then 
*  call a function to convert the old data structure into
*  the new structure.
*
* \param filename of file that contains the old file format
* \param settings of current rootfly experiment
* \return returns 0 on success
*/

int Experiment::LoadOld(const char *filename, Settings *parameters)
{
  
  OldData::RootflyData oldDataStructure;
  
  bool success;
  
  m_image_file_format.SetToBartz();
  success = oldDataStructure.Load((CString)filename);
  CString msg = "";
  int status = 0;
  
  if (success == true)
  {
    status = ConvertDataStructure(oldDataStructure, parameters, (CString)filename);
    if (status != 0)  return status;
    
    //    if(!TestingData::compareDataStructure(oldDataStructure,this))
    //    {
    //      
    //      return -1;
    //    }
    
  }
  else
  {
    return ERRORS::RFY_LOADERROR_FILECORRUPT;
  }
  
  
  return 0;
}


/**
@class Settings

  @brief This class holds all of the permanent user preferences that persist between runs of the
  application.  These data are stored in the Windows registry.
  
    This class contains user preferences for displaying such as the GUI checkboxes, 
    the color, and the last screen location of the application window.
*/

// this should be the authoritative source for the version information
const CString Settings::m_version = "Rootfly_21";

// initialize root drawing colors (unselected = red, selected = green, dead = black)
const COLORREF Settings::m_default_root_drawing_color = RGB(255, 0, 0);
const COLORREF Settings::m_default_root_drawing_color_selected = RGB(0, 255, 0);
const COLORREF Settings::m_default_root_drawing_color_dead = RGB(0, 0, 0);
const COLORREF Settings::m_default_root_drawing_color_dead_selected = RGB(200, 200, 200);

Settings::Settings()
  : m_root_drawing_color ( GetDefaultRootDrawingColorLive() ),
    m_root_drawing_color_selected (GetDefaultRootDrawingColorLiveSelected() ),
    m_root_drawing_color_dead (GetDefaultRootDrawingColorDead() ),
    m_root_drawing_color_dead_selected (GetDefaultRootDrawingColorDeadSelected() ),
    m_condition_default_index ( 0 ),
    m_show_roots (true), 
    m_show_root_ids (true), 
    m_show_dead_roots (true),
    m_show_diameters (true),
    m_check_diameter (false),
    m_check_length (false), 
    m_check_vertical (false),
    m_check_post_process (false),
    m_diameter_value (0),
    m_length_value (0),
    m_degree_value (0),
    m_new_root_color(RootInstance::COLOR_WHITE),
    m_new_root_is_alive(true)
{
  m_condition_vector.clear();
}

Settings::~Settings()
{
}

void Settings::SaveUserPreferences() const
{
  CString condition_name;
  int condition_size = m_condition_vector.size();
  Registry reg_instance("Settings");
  
  reg_instance.SaveStringToReg("Version", m_version);
  reg_instance.SaveStringToReg("FileDir", m_default_file_dir);
  reg_instance.SaveIntToReg("ConditionCodeDefaultIndex", m_condition_default_index);
  
  reg_instance.SaveIntToReg("RootIdEnabled",GetShowRootID());
  reg_instance.SaveIntToReg("DeadRootsEnabled",GetShowDeadRoots());
  reg_instance.SaveIntToReg("ShowDiametersEnabled",GetShowDiameters());
  
  reg_instance.SaveIntToReg("RootDrawingColor", GetRootDrawingColorLive());
  reg_instance.SaveIntToReg("RootDrawingColorSelected", GetRootDrawingColorLiveSelected());
  reg_instance.SaveIntToReg("RootDrawingColorDead", GetRootDrawingColorDead());
  reg_instance.SaveIntToReg("RootDrawingColorDeadSelected", GetRootDrawingColorDeadSelected());
  
  CString auto_detect_cstring;
  
  reg_instance.SaveIntToReg("CheckDiameter", m_check_diameter);
  auto_detect_cstring.Format("%f",m_diameter_value);
  reg_instance.SaveStringToReg("DiameterValue", auto_detect_cstring);
  
  reg_instance.SaveIntToReg("CheckLength", m_check_length);
  auto_detect_cstring.Format("%f",m_length_value);
  reg_instance.SaveStringToReg("LengthValue", auto_detect_cstring);
  
  reg_instance.SaveIntToReg("CheckVertical", m_check_vertical);
  auto_detect_cstring.Format("%f",m_degree_value);
  reg_instance.SaveStringToReg("DegreeValue", auto_detect_cstring);

  auto_detect_cstring.Format("%d", m_last_app_window_location.left);
  reg_instance.SaveStringToReg("LastWinLocationLeft", auto_detect_cstring);

  auto_detect_cstring.Format("%d", m_last_app_window_location.top);
  reg_instance.SaveStringToReg("LastWinLocationTop", auto_detect_cstring);
  
  auto_detect_cstring.Format("%d", m_last_app_window_location.right);
  reg_instance.SaveStringToReg("LastWinLocationRight", auto_detect_cstring);

  auto_detect_cstring.Format("%d", m_last_app_window_location.bottom);
  reg_instance.SaveStringToReg("LastWinLocationBottom", auto_detect_cstring);

  int diam_scaling = (int) ((m_diameter_scaling_factor * 10.0) + 0.5);  // store the first decimal point
  reg_instance.SaveIntToReg("DiameterScalingFactor", diam_scaling);

  reg_instance.SaveIntToReg("ConditionSize", condition_size);
  for(int i = 0; i < condition_size; i++)
  {
    condition_name.Format("Condition_%d",i);
    reg_instance.SaveStringToReg(condition_name, m_condition_vector[i]);
  }
  
  reg_instance.SaveIntToReg("NewRootColor", m_new_root_color);
  reg_instance.SaveIntToReg("NewRootIsAlive", m_new_root_is_alive);
}

// Sets 'color' to the color found in registry with key 'str'.
// If key not found, then does not set 'color'.
void GetColorFromReg(const Registry& reg, const CString& str, COLORREF* color)
{
  COLORREF col;
  col = reg.GetIntFromReg(str);
  if (col != -1)  *color = col;
}

void Settings::ReadUserPreferences()
{
  CString condition_name;
  const Registry reg_instance("Settings");
  m_default_file_dir = reg_instance.GetStringFromReg("FileDir");
  m_condition_default_index = reg_instance.GetIntFromReg("ConditionCodeDefaultIndex");
  if (m_condition_default_index < 0)  m_condition_default_index = 0;
  int condition_size = reg_instance.GetIntFromReg("ConditionSize");
  
  // all member bools below are set to true if registry key doesn't exist
  // making the default for all the checkboxes to be turned on
  m_show_root_ids = reg_instance.GetIntFromReg("RootIdEnabled") != 0 ? true:false;
  m_show_dead_roots = reg_instance.GetIntFromReg("DeadRootsEnabled") != 0 ?true:false;
  m_show_diameters = reg_instance.GetIntFromReg("ShowDiametersEnabled") != 0 ?true:false;
  
  GetColorFromReg(reg_instance, "RootDrawingColor", &m_root_drawing_color);
  GetColorFromReg(reg_instance, "RootDrawingColorSelected", &m_root_drawing_color_selected);
  GetColorFromReg(reg_instance, "RootDrawingColorDead", &m_root_drawing_color_dead);
  GetColorFromReg(reg_instance, "RootDrawingColorDeadSelected", &m_root_drawing_color_dead_selected);
  
  CString auto_detect_cstring;
  
  m_check_diameter = reg_instance.GetIntFromReg("CheckDiameter") != 1 ? 0:1;
  auto_detect_cstring = reg_instance.GetStringFromReg("DiameterValue");
  m_diameter_value = atof((const char *)(LPCTSTR)auto_detect_cstring);
  
  m_check_length = reg_instance.GetIntFromReg("CheckLength") != 1 ? 0:1;
  auto_detect_cstring = reg_instance.GetStringFromReg("LengthValue");
  m_length_value = atof((const char *)(LPCTSTR)auto_detect_cstring);
  
  m_check_vertical = reg_instance.GetIntFromReg("CheckVertical") != 1 ? 0:1;
  auto_detect_cstring = reg_instance.GetStringFromReg("DegreeValue");
  m_degree_value = atof((const char *)(LPCTSTR)auto_detect_cstring);

  auto_detect_cstring = reg_instance.GetStringFromReg("LastWinLocationLeft");
  m_last_app_window_location.left = atoi((const char *)(LPCTSTR)auto_detect_cstring);

  auto_detect_cstring = reg_instance.GetStringFromReg("LastWinLocationTop");
  m_last_app_window_location.top = atoi((const char *)(LPCTSTR)auto_detect_cstring);

  auto_detect_cstring = reg_instance.GetStringFromReg("LastWinLocationRight");
  m_last_app_window_location.right = atoi((const char *)(LPCTSTR)auto_detect_cstring);

  auto_detect_cstring = reg_instance.GetStringFromReg("LastWinLocationBottom");
  m_last_app_window_location.bottom = atoi((const char *)(LPCTSTR)auto_detect_cstring);

  m_diameter_scaling_factor = reg_instance.GetIntFromReg("DiameterScalingFactor") / 10.0;
  if (m_diameter_scaling_factor < 0) m_diameter_scaling_factor = 3.0;

  m_condition_vector.clear();
  for(int i = 0; i < condition_size; i++)
  {
    condition_name.Format("Condition_%d",i);
    CString code = reg_instance.GetStringFromReg(condition_name);
    
    // Check to make sure that code is valid.  This should never be a problem unless
    // the user manually edits the registry, because ManageConditionCodesDlg::AddNewCode() 
    // checks for validity whenever the user enters a new code.
    // Nevertheless, this is still worth checking because an invalid 
    // condition code will prevent .rfy files from being loaded properly (fgets
    // skips over leading whitespace), which
    // can cause severe data loss for the user.  Better not to risk it.
    if (code.GetLength() > 0 && code[0] != ' ' && code[0] != '\t')
    {
      m_condition_vector.push_back(code);
    }
  }
  
  int val = reg_instance.GetIntFromReg("NewRootColor");
  m_new_root_color = static_cast<RootInstance::Color>( (val < 0) ? 0 : val);
  m_new_root_is_alive = reg_instance.GetIntFromReg("NewRootIsAlive") != 0 ? true:false;;
}

CurrentImageProperties::CurrentImageProperties()
  : m_experiment_name(),
    m_current_image_filename(),
    m_session_date(),
    m_operator_initials(),
    m_tube(),
    m_window(-1),
    m_session(-1)
{
}

ApplicationData::ApplicationData()
  : m_modified(false),
    m_filename(),
    m_curr(),
    m_main_image(),
    m_temp_image()
{
}

ApplicationData::~ApplicationData()
{ 
}

// completely reloads the image buffers,
void ApplicationData::ReloadImageBuffers(const CString& dir, const ImageFileFormat& file_format,SessionCollapser &sess_coll)
{
  // Refresh ImageManager cache
  // Initialize should really be called somewhere else.  Maybe in OnFileNew and OnFileLoad after the image directory, file format, and session collapser have been set.
  m_image_mgr.Initialize( dir, file_format, sess_coll );
  m_image_mgr.SetCurrentTWS( m_curr.m_tube, m_curr.m_window, m_curr.m_session );
}

// !!! SessionCollapser:GetRemappedIndex contains an assert for bounds checking. Could fail more gracefully
// !!! SessionCollapser::RemapSessionIndices if statement at line 1975 "if(g_collapse_sessions == false) return;"
//     could possibly be placed at the beginning of the function

//initializes for non-collapsing
SessionCollapser::SessionCollapser()
{
  int i;
  for(i = -2; i <= 2; i++)
  {
    m_indices[i+2] = i;
  }
}

//grab needed data
void SessionCollapser::Init(Experiment *exp, ApplicationData * appdata)
{
  m_exp = exp;
  m_appdata = appdata;
}

/*return desired index. Range should be between -2 and + 2 inclusive. It will either
return the same value as passed in, or will return a index to an image that exists.
Example: if remapping is enabled, passing in a -2 would return the second image to 
the left of the center image, as opposed to the second session to the left of center.
*/
int SessionCollapser::GetRemappedIndex(int unmapped_index) const
{
  if(unmapped_index < -2 || unmapped_index > 2)
  {
    assert(0);
    return 0;
  }
  else return m_indices[unmapped_index+2];
}

//this is where the work is done
void SessionCollapser::RemapSessionIndices()
{
  //grab current tube/window/session
  TubeName tubename = m_appdata->GetCurrentTubeName();
  const Tube *tube = m_exp->GetTubeByName(tubename);
  int window = m_appdata->GetCurrentWindow();
  int session = m_appdata->GetCurrentSession();
  
  int i,j;
  //reinitialize to remove any wonky values
  for(i = -2; i <= 2; i++)
  {
    m_indices[i+2] = i;
  }
  if(g_collapse_sessions == false) return;  // don't do anything unless desired
  
  //start from center
  m_indices[2] = 0;
  
  //remap -1 and -2
  int marker = session -1;
  for(i = -1; i >= -2; i--)
  {
    //find images to the left that exist and mark the first two
    for(j = marker; j > 0; j--)
    {
      if(tube->GetImageExists(window, j))
      {
        m_indices[i+2] = j - session;  //mark image
        marker = j -1;
        break;
      }
      else continue;  //keep looking
    }
  }
  if(m_indices[0] >= m_indices[1]) m_indices[0] = m_indices[1] - 1;  //only one image found
  
  //remap 1 and 2
  marker = session + 1;
  for(i = 1; i <= 2; i++)
  {
    int max = m_exp->GetMaxSessions();
    if(max < 1) max = 1;
    //find images that exist to the right and mark the first two
    for(j = marker; j <= max; j++)
    {
      if(tube->GetImageExists(window, j))
      {
        m_indices[i+2] = j - session;  //mark image
        marker = j + 1;
        break;
      }
      if(j > 100 + marker) break;
      else continue;
    }
  }
  if(m_indices[4] <= m_indices[3]) m_indices[4] = m_indices[3] + 1;  //only one image found
  
}

SessionCollapser::~SessionCollapser()
{
}


/**
* Default Constructor
*/
RootflyData::RootflyData() 
{
  m_sess_coll.Init(&m_exp, &m_appdata);
}

/**
* Default Destructor
*/
RootflyData::~RootflyData()
{
}


void Registry::SaveStringToReg(const CString& name, const CString& data)
{
  int res = pApp->WriteProfileString(m_section, name, data);
  assert( res != 0 );
}

CString Registry::GetStringFromReg(const CString& name) const
{
  // if entry not found in registry, return empty string
  return pApp->GetProfileString(m_section, name, "");
}

void Registry::SaveIntToReg(const CString& name, const int data)
{
  pApp->WriteProfileInt(m_section, name, data);
}

int Registry::GetIntFromReg(const CString& name) const
{
  // if entry not found in registry, returns -1
  return pApp->GetProfileInt(m_section, name, -1);
}





///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// ///////////////////////////////////////////////////////////////////////// //
// /////////////                                            //////////////// //
// /////////////            FILE I/O Group Code             //////////////// //
// /////////////                                            //////////////// //
// ///////////////////////////////////////////////////////////////////////// //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////


// returns true upon success, false otherwise
bool LoadConditionCode( FILE* fpt, CString* out )
{
  enum { NCHARS = 1000 };
  char          tempBuf[NCHARS];
  // Note:  If the condition code is the empty string, then this code will fail.
  // It will ignore the empty string and instead treat the next line as the condition code.
  // It also eats up leading whitespace in a condition code.
  // Our solution for now is not to allow empty string condition codes or codes with
  // leading whitespace.
  char *status;
  status = fgets ( tempBuf, NCHARS, fpt);
  tempBuf[strlen(tempBuf)-1] = '\0';
  *out = tempBuf;
  
  //  {  // temp
  //    assert(strcmp(*out, "NO_CODE") == 0);
  //  }
  
  return (status != NULL);
} 

/**
* Loads a RootInstance of RPoints and RDiameters into the current data structure.
*
* The file pointer will be located at the beginning of a RootInstance
* and the function will load all points and diameters associated with the 
* RootInstance
*
* @author Hamilton Evans
* @param fpt pointer to an open file handle.
* @return Success (0) or Failure (-1)
*/
int RootInstance::Load(FILE *fpt)
{
  if(!fpt) return -1;
  int           i, num_points, num_diameters, color, x, y, window, session_num, liveordead;
  double        radius = 0.0;
  CString       user_code;
  RPoint        *new_point;
  RDiameter     *new_diameter;
  
  
  fscanf(fpt,"%d %d %d %d %d\n", &session_num, &num_points, &num_diameters, &color, &liveordead); 
  
  // checks if we were able to read the condition code
  if (!LoadConditionCode( fpt, &m_condition_code ))
  {
    fclose(fpt);
    return -1;
  }
  
  
  // !!! We can load these values directly with = or in the fscanf line -- ADP
  SetColor((RootInstance::Color)color);
  
  m_is_alive = (liveordead == 1);
  
  SetSessionNumber(session_num);
  
  // Load the points for the instance
  for(i=0;i<num_points;i++)
  {
    new_point = (RPoint*)CreatePoint();
    
    fscanf(fpt,"%d %d %d\n",&x,&y,&window); 
    new_point->x = x;
    new_point->y = y;
    new_point->window = window;
  }
  
  // Load the diameters for the instance
  for(i=0;i<num_diameters;i++)
  {
    new_diameter = (RDiameter*)CreateDiameter();
    
    fscanf(fpt,"%d %d %d %lf\n",&x,&y,&window,&radius);
    new_diameter->radius = radius;
    new_diameter->r_point.x = x;
    new_diameter->r_point.y = y;
    new_diameter->r_point.window = window;
  }
  
  return 0;
}

/**
* Saves information from the current datastructure
* @param fp pointer to an open file handle.
* @return Success (0) or Failure (-1)
*/

// !!! We can remove all the 4 copy variables below and print them directly
// !!! Bools can print and scan as ints just fine (if int is 0 and 1 only?)
int RootInstance::Save(FILE *fp) const
{
  g_debug_logfile.Print("RootInstance::Save() 1\n");
  int num_points, num_diams, color, status, i;
  
  num_points  = m_points.size();
  num_diams   = m_diameters.size();
  color       = m_color;
  status      = (m_is_alive) ? 1 : 0;
  
  {
    // Fix condition code if it happens not to conform to expected format.
    // Condition codes should not be empty strings, or begin with leading whitespace.
    // This is being checked in ManageConditionCodesDlg::AddNewCode, so there is 
    // no way that this should happen.  But for some reason people are reporting it as
    // happening.  This is an ugly fix that hopefully we can get rid of one day.  
    // And we really should fix the Load function, but that will require changing the
    // fopen function to load .rfy files in binary format so that we can check for
    // the empty string.  See bug 2617346.  -- STB 4/15/09
    // Note:  In UC_CreateRoot::Execute(), I moved SetConditionCode() to just below the
    //    assignment operator.  This fixed at least one place where the condition code
    //    was being set to the empty string.  Perhaps it was the only place(?).  -- STB 7/2/09
    CString new_code;
    if (strlen(m_condition_code)==0)
    {
      new_code = "NO_CODE";
    }
    else if (m_condition_code[0] == ' ')
    {
      new_code = m_condition_code;
      new_code.SetAt(0, '_');
    }
    if (strlen(new_code)>0)
    {
      assert( 0 );
      const_cast<CString*>(&m_condition_code)->operator=( new_code );
    }
  }
  
  g_debug_logfile.Print("RootInstance::Save() 2\n");
  //Note, instead of the date this now saves the session number as an int only.
  fprintf(fp, "%d %d %d %d %d\n%s\n", m_session_number, num_points, num_diams, color, status, m_condition_code);
  
  //Put the points in the file
  g_debug_logfile.Print("RootInstance::Save() 3 %d <= %d\n", num_points, m_points.size());
  assert(num_points == m_points.size());  // this assert is almost always pointless -- ADP
  for(i=0; i<num_points; i++) 
  {
    g_debug_logfile.Print("RootInstance::Save() 3b %d\n", i);
    if(fprintf(fp, "%d %d %d\n",m_points[i].x, m_points[i].y, m_points[i].window) < 0)
    {
      return -1;
    }
    
  }
  g_debug_logfile.Print("RootInstance::Save() 4 %d <= %d\n", num_diams, m_diameters.size());
  assert(num_diams == m_diameters.size());
  //Put the diameters in the file
  for(i=0; i<num_diams; i++) {
    g_debug_logfile.Print("RootInstance::Save() 4b %d\n", i);
    if(fprintf(fp, "%d %d %d %f\n",   m_diameters[i].r_point.x, 
      m_diameters[i].r_point.y, 
      m_diameters[i].r_point.window, 
      m_diameters[i].radius) < 0)
    {
      return -1;
    }
    
  }
  g_debug_logfile.Print("RootInstance::Save() 5\n");
  
  
  return 0;
}


/**
* Loads a Root of RootInstances into the current data structure.
*
* The file pointer will be located at the beginning of a Root and the 
* function will load all RootInstances associated with the Root.
*
* @author Hamilton Evans
* @param fpt pointer to an open file handle.
* @return Success (0) or Failure (-1)
*/

// Probably should assert num_sessions == m_instances.size() at the end -- ADP
int Root::Load(FILE *fpt)
{
  if(!fpt) return -1;
  g_debug_logfile.Print("Root::Load() 1\n");
  
  int           i, first_session, num_sessions, root_id, window_id, primary_window;
  RootInstance  *new_instance;
  
  fscanf(fpt,"\n%d %d %d %d %d\n",&first_session,&num_sessions, &root_id, &window_id, &primary_window);
  
  g_debug_logfile.Print("Root::Load() 2 "
    "first_session=%d num_sessions=%d root_id=%d window_id=%d primary_window=%d\n",
    first_session, num_sessions, root_id, window_id, primary_window);
  
  g_debug_logfile.Print("Root::Load() 2a pw=%d %d %d\n", primary_window, primary_window > 9999, (first_session > 9999 || num_sessions > 9999  || root_id > 9999  || window_id > 9999 || primary_window > 9999));
  if (first_session > 9999 || num_sessions > 9999  || root_id > 9999  || window_id > 9999 || primary_window > 9999)
  {
//   would be nice to get the line number here; could rewind to beginning of file, then search for '\n' until we reach fpos again -- STB
//    fpos_t fpos;
//    fgetpos(fpt, &fpos);
    AfxMessageBox("Error:  Root has excessively large session or window number.  There is probably an empty condition code in this file, or the file is corrupted.\n");  
    return -1;
  }
  
  m_id = root_id;
  m_primary_window = primary_window;
  m_window_id = window_id;
  
  /// Begin numbering sessions at first_session, end when the number
  /// of sessions for a specific root are completed (i.e. - num_sessions).
  for(i=0;i<num_sessions;i++)
  {
    new_instance = (RootInstance*)CreateRootInstance(first_session+i);
    new_instance->Load(fpt);
  }
  
  g_debug_logfile.Print("Root::Load() 3\n");
  return 0;
}

/**
* Saves information from the current datastructure
* @param fp pointer to an open file handle.
* @return Success (0) or Failure (-1)
*/
int Root::Save(FILE *fp) const
{
  g_debug_logfile.Print("Root::Save() 1\n");
  int i, num_sessions=0, first_session=0;
  bool run=false;
  
  num_sessions  = m_instances.size();
  
  g_debug_logfile.Print("Root::Save() 2\n");
  //If there are any instances defined, grab the session number of the first root instance.
  if(num_sessions > 0) {
    first_session = m_instances.at(0).GetSessionNumber();
  }
  
  g_debug_logfile.Print("Root::Save() 3 (%d %d %d %d %d)\n", first_session, num_sessions, m_id, m_window_id, m_primary_window);
  if(fprintf(fp, "%d %d %d %d %d\n", first_session, num_sessions, m_id, m_window_id, m_primary_window) < 0) 
  {
    return -1;
  }
  
  
  g_debug_logfile.Print("Root::Save() 4\n");
  for(i=0; i<num_sessions; i++) 
  {
    g_debug_logfile.Print("Root::Save() 4b %d\n", i);
    run=true;
    if((m_instances[i]).Save(fp) == -1) 
    {
      return -1;
    }
    fprintf(fp, "\n");
  }
  
  g_debug_logfile.Print("Root::Save() 5\n");
  if(!run) 
  {
    fprintf(fp,"\n");
  }
  
  g_debug_logfile.Print("Root::Save() 6\n");
  return 0;
}


/**
* Loads a Tube of Roots into the current data structure.
*
* The file pointer will be located at the beginning of a Tube and the 
* function will load all Roots associated with the Tube.
*
* @author Hamilton Evans
* @param fpt pointer to an open file handle.
* @return Success (0) or Failure (-1)
*/
int Tube::Load(FILE *fpt)
{
  if(!fpt) return -1;
  g_debug_logfile.Print("Tube::Load() 1\n");
  
  int           i, num_roots, size, temp_window, temp_session, sizeImageAnnotated;
  Root          *new_root;
  ImageData temp;
  ImageData imageAnnotated;
  
  // loads size of imagescanned vector
  fscanf(fpt,"%d\n", &size);
  
  g_debug_logfile.Print("Tube::Load() 2 size=%d\n", size);
  if (size > 9999)
  {
    AfxMessageBox("Error:  Tube has an excessively large number of roots.\n");  
    return -1;
  }
  
  // load imagescanned vector
  for(i = 0; i < size; i++)
  {
    
    fscanf(fpt, "%d %d\n", &temp_window, &temp_session);
    temp.window = temp_window;
    temp.session = temp_session;
    m_image_scanned.push_back(temp);
  }
  
  // START new file version starting with 1.8.25
  // loads size of imageAnnotated vector
  fscanf(fpt,"%d\n", &sizeImageAnnotated);
  
  g_debug_logfile.Print("Tube::Load() 3 sizeImageAnnotated=%d\n", sizeImageAnnotated);
  
  // load imageAnnotated vector
  for(i = 0; i < sizeImageAnnotated; i++)
  {
    
    fscanf(fpt, "%d %d\n", &temp_window, &temp_session);
    imageAnnotated.window = temp_window;
    imageAnnotated.session = temp_session;
    m_image_annotated.push_back(imageAnnotated);
  }
  // END new file version starting with 1.8.25
  
  /// Read the file to get the number of roots and the tag for the tube.
  fscanf(fpt,"%d\n", &num_roots);                           
  
  g_debug_logfile.Print("Tube::Load() 4 numroots=%d\n", num_roots);
  
  for(i=0;i<num_roots;i++)
  {
    new_root = (Root*) CreateRoot();
    int ret = new_root->Load(fpt);
    if (ret < 0)  return -1;
  }
  
  g_debug_logfile.Print("Tube::Load() 5 m_root.size()=%d\n", m_root.size());
  
  { // set 'm_next_root_id'
    int max_root_id = 0;  // This is zero, b/c root ids are one-based indices (see +1 below)
    for (i=0 ; i<m_root.size() ; i++)
    {
      int id = m_root[i].GetID();
      if (id >= max_root_id)  max_root_id = id;
    }
    m_next_root_id = max_root_id + 1;
  }
  
  g_debug_logfile.Print("Tube::Load() 6\n");
  
  return 0;
}

/**
* Saves information from the current datastructure
* @param fp pointer to an open file handle.
* @return Success (0) or Failure (-1)
*/
int Tube::Save(FILE *fp) const
{
  if(!fp) return -1;
  g_debug_logfile.Print("Tube::Save() 1\n");
  int num_roots, i;
  
  num_roots = m_root.size();
  g_debug_logfile.Print("Tube::Save() 2 %d\n", num_roots);
  
  
  if(fprintf(fp, "%s\n", m_tag) < 0) 
  {
    return -1;
  }
  g_debug_logfile.Print("Tube::Save() 3\n");
  
  // save imagescanned size
  if(fprintf(fp, "%d\n", m_image_scanned.size()) < 0)
  {
    return -1;
  }
  g_debug_logfile.Print("Tube::Save() 4\n");
  // save imagescanned vector
  for(i = 0; i < m_image_scanned.size(); i++)
  {
    g_debug_logfile.Print("Tube::Save() 4b %d\n", i);
    if(fprintf(fp, "%d %d\n", m_image_scanned[i].window,
      m_image_scanned[i].session) < 0)
    {
      return -1;  // problems with saving
    }
    
  }
  g_debug_logfile.Print("Tube::Save() 5\n");
  
  // START new file version starting with 1.8.25
  // save imageAnnotated size
  if(fprintf(fp, "%d\n", m_image_annotated.size()) < 0)
  {
    return -1;
  }
  g_debug_logfile.Print("Tube::Save() 6\n");
  // save imageAnnotated vector
  for(i = 0; i < m_image_annotated.size(); i++)
  {
    g_debug_logfile.Print("Tube::Save() 6b %d\n", i);
    if(fprintf(fp, "%d %d\n", m_image_annotated[i].window,
      m_image_annotated[i].session) < 0)
    {
      return -1;  // problems with saving
    }
    
  }
  // END new file version starting with 1.8.25
  
  
  g_debug_logfile.Print("Tube::Save() 7\n");
  if(fprintf(fp, "%d\n", num_roots) < 0) 
  {
    return -1;
  }
  
  
  g_debug_logfile.Print("Tube::Save() 8 %d <= %d\n", num_roots, m_root.size());
  assert(num_roots == m_root.size());
  for(i=0; i<num_roots; i++) {
    g_debug_logfile.Print("Tube::Save() 8b %d\n", i);
    if((m_root[i]).Save(fp) == -1) 
    {
      return -1;
    }
  }
  g_debug_logfile.Print("Tube::Save() 9\n");
  return 0;
}
